import communications.HDLC_communication
import struct
import time
import common.helpers
from common import helpers
from communications import HDLC_communication

LTC5800IPR_PROTOCOL_VERSION = 4

RC_OK               = 0
RC_INVALID_COMMAND  = 1
RC_INVALID_ARGUMENT = 2
RC_END_OF_LIST      = 11
RC_NO_RESOURCES     = 12
RC_IN_PROGRESS      = 13
RC_NACK             = 14
RC_WRITE_FAIL       = 15
RC_VALIDATION_ERROR = 16
RC_INV_STATE        = 17
RC_NOT_FOUND        = 18
RC_UNSUPPORTED      = 19

MOTE_STATE_LOST         = 0
MOTE_STATE_NEGOTIATING  = 1
MOTE_STATE_OPERATIONAL  = 4

SMIP_MAX_PAYLOAD_TO_MANAGER = 90


class class_Dongle_Communication(HDLC_communication.class_HDLC_Communication):
    def __init__(self,
                 comm_port,
                 callback_command_resp=None,            # response for commands
                 callback_notif_event=None,             # notification.events
                 callback_notif_log=None,               # notification.logs
                 callback_notif_data=None,              # notifications.data - here data from mote should be processed
                 callback_notif_ipData=None,            # notification.ipData - here data from mote using ipv6 should be processed
                 callback_notif_healthReport=None,      # notification.health RawDataSensors
                 callback_MgrHello=None,                # when manager disconnects from us and waiting for ClientHello message (is_active_session -> False)
                 callback_HelloResponse=None,           # when manager connects us and waiting for further messages from us (is_active_session -> True)
                 callback_serial_comm_open=None,        # serial COM port is now open
                 callback_serial_comm_close=None):      # serial COM port is now closed
        HDLC_communication.class_HDLC_Communication.__init__(
            self=self,
            comm_port=comm_port,
            callback_process_message=self.__process_message__,
            callback_serial_comm_open=self.__on_serial_comm_open__,
            callback_serial_comm_close=self.__on_serial_comm_close__
        )
        self.callback_command_resp = callback_command_resp
        self.callback_serial_comm_open = callback_serial_comm_open
        self.callback_serial_comm_close = callback_serial_comm_close
        self.callback_notif_event = callback_notif_event
        self.callback_notif_log = callback_notif_log
        self.callback_notif_data = callback_notif_data
        self.callback_notif_ipData = callback_notif_ipData
        self.callback_notif_healthReport = callback_notif_healthReport
        self.callback_MgrHello = callback_MgrHello
        self.callback_HelloResponse = callback_HelloResponse
        self.in_active_session = False
        self.MgrHelloCounter = 0    # up counter to decide when to send clientHello again
        self.mgrSeqNo = 0   # manager sequence number as received in HelloResponse packet, and by acknowledged messages that the manager sends us
        self.cliSeqNo = 0   # iComox sequence number sent in Hello message, and in iComox request messages
        self.seqNumber = 0  # contains the sequence number of the packet that was acknowledged. write_to_manager() checks it when it waits for response for messages that need to get ACK
        # manage the mechanism to send messages that requires returned ACK responses, again and again
        self.write_attempts = 0
        self.latest_time_for_ack_to_arrive = 0
        self.msg_requires_ack = bytearray()

    def __on_serial_comm_open__(self, comm):
        helpers.OUT("LTC5800IPR.__on_serial_comm_open__")
        self.in_active_session = False
        self.MgrHelloCounter = 0
        if callable(self.callback_serial_comm_open):
            self.callback_serial_comm_open(self)
        self.send_ClientHello()

    def __on_serial_comm_close__(self, comm):
        helpers.OUT("LTC5800IPR.__on_serial_comm_close__")
        self.in_active_session = False
        self.MgrHelloCounter = 0
        if callable(self.callback_serial_comm_close):
            self.callback_serial_comm_close(self)

    def __process_message__(self, comm, msg):
        #   helpers.OUT("LTC5800IPR.__process_message__")
        if len(msg) < 4:
            return
        control, packetType, seqNumber, payloadLen = struct.unpack("BBBB", msg[:4])

        if (control & 1) != 0:      # iComox received ack packet
            #   helpers.OUT("LTC5800IPR: ACK packet received")
            if (control & 2) != 0:  # iComox previously requested ACK
                if seqNumber == self.seqNumber:
                    return  # ignore the packet
                else:
                    self.seqNumber = seqNumber  # store the new incoming sequence number and process the message
            else:                   # iComox previously did not request ACK
                pass
        else:                       # iComox received data packet (probably notification)
            if (control & 2) != 0:  # manager requires ACK
                self.write_to_manager(packetType=packetType, AckPacket=True, serviceTypeAck=True)  # send ACK to the received message
            else:                   # manager requires no ACK
                pass
        # if (control & 1) != 0:      # dongle application received ack packet
        #     #   helpers.OUT("LTC5800IPR: ACK packet received")
        #     if (control & 2) != 0:  # dongle application previously requested ACK
        #         if seqNumber != self.cliSeqNo:
        #             self.send_ClientHello(cliSeqNo=0xFF)
        #             return
        #         else:
        #             pass
        #     else:                   # dongle application previously did not request ACK
        #         pass
        # else:                       # dongle application received data packet (probably notification)
        #     if (control & 2) != 0:  # manager requires ACK
        #         self.write_to_manager(packetType=packetType, AckPacket=True, serviceTypeAck=True)  # send ACK to the received message
        #         if seqNumber == self.seqNumber: # if duplicated packet
        #             return  # ignore the packet
        #         else:
        #             self.seqNumber = seqNumber  # store the new incoming sequence number and process the message
        #     else:                   # manager requires no ACK
        #         pass

        payload = msg[4:]

        if packetType == 0x14:  # notifications which can be events
            #   helpers.OUT("LTC5800IPR.notification received")
            if (payloadLen > 124) or (len(msg) != payloadLen + 4):
                return
            if len(payload) < 1:
                return
            notifType = payload[0]
            if notifType == 1:       # Event notification
                if len(payload) < 6:
                    return
                if callable(self.callback_notif_event):
                    eventId, eventType = struct.unpack_from(">LB", payload[:6], 1)
                    eventData = payload[6:]
                    self.callback_notif_event(self, eventId, eventType, eventData)

            elif notifType == 2:     # Log notification
                if len(payload) < 9:
                    return
                if callable(self.callback_notif_log):
                    macAddress = payload[1:9]
                    logMsg = payload[9:]
                    self.callback_notif_log(self, macAddress, logMsg)

            elif notifType == 4:     # Data payload notification
                if len(payload) < 25:
                    return
                if callable(self.callback_notif_data):
                    timestamp = payload[1:13]
                    macAddress = payload[13:21]
                    srcPort, dstPort = struct.unpack_from(">HH", payload[:25], 21)
                    data = payload[25:]
                    self.callback_notif_data(self, timestamp, macAddress, srcPort, dstPort, data)

            elif notifType == 5:     # 6lowpan packet notification
                if len(payload) < 21:
                   return
                if callable(self.callback_notif_ipData):
                    timestamp = payload[1:13]
                    macAddress = payload[13:21]
                    data = payload[21:]
                    self.callback_notif_ipData(self, timestamp, macAddress, data)

            elif notifType == 6:     # Health report notification
                if len(payload) < 9:
                    return
                if callable(self.callback_notif_healthReport):
                    macAddress = payload[1:9]
                    data = payload[9:]
                    self.callback_notif_healthReport(self, macAddress, data)

            else:
                return

        elif packetType == 2:   # Hello response
            helpers.OUT("LTC5800IPR.HelloResponse")
            if len(payload) < 5:
                return
            RC, version, self.mgrSeqNo, self.cliSeqNo, mode = struct.unpack("BBBBB", payload[:5])
            if RC == 0:
                self.in_active_session = True
                self.MgrHelloCounter = 0
                if callable(self.callback_HelloResponse):
                    self.callback_HelloResponse(self, RC, version, mode)
            else:
                self.send_ClientHello(self)

        elif packetType == 3:   # MgrHello
            helpers.OUT("LTC5800IPR.MgrHello")
            self.in_active_session = False
            self.MgrHelloCounter += 1
            if callable(self.callback_MgrHello):
                self.callback_MgrHello(self, self.MgrHelloCounter)
            #   self.send_ClientHello()

        else:
            if (len(payload) == 0) or (payloadLen > 124) or (len(msg) != payloadLen + 4):   # The RC field is not included in the payloadLen????
                helpers.OUT("__process_message__(): Illegal payload field, or payload does not contain RC field")
                return      # illegal payloadLen field, or payload does not contain the RC field

            if callable(self.callback_command_resp):
                RC = payload[0]
                if len(payload) < 2:
                    data = bytearray()
                else:
                    data = payload[1:]
                self.callback_command_resp(self, packetType, RC, data)

    def write_service_type_ack_msg_to_manager(self):
        if self.write_attempts > 0:
            self.write_attempts -= 1
            self.latest_time_for_ack_to_arrive = time.monotonic() + 0.2   # 200 msec before trying again
            result = self.write_hdlc_msg(msg=self.msg_requires_ack)
            if not result:
                self.write_attempts = 0
                self.close()
        else:
            self.send_ClientHello(cliSeqNo=0)
            result = False
        return result

    def write_to_manager(self, packetType, payload=bytearray(), AckPacket=False, serviceTypeAck=True):
        Control = 0
        if AckPacket:
            Control |= 1
        if serviceTypeAck:
            Control |= 2

        if serviceTypeAck:
            if AckPacket:
                seqNo = self.mgrSeqNo
            else:
                self.cliSeqNo = (self.cliSeqNo + 1) % 256
                seqNo = self.cliSeqNo
        else:
            seqNo = 0

        msg = bytearray([Control, packetType, seqNo, len(payload)]) + payload

        if serviceTypeAck:
            self.msg_requires_ack = msg
            self.write_attempts = 1
            return self.write_service_type_ack_msg_to_manager()
        else:
            return self.write_hdlc_msg(msg=msg)

    def send_ClientHello(self, cliSeqNo=None):
        helpers.OUT("ClientHello")
        if cliSeqNo is not None:
            self.cliSeqNo = cliSeqNo
        payload = bytearray(struct.pack("BBB", LTC5800IPR_PROTOCOL_VERSION, self.cliSeqNo, 0))

        return self.write_to_manager(packetType=0x01, payload=payload, serviceTypeAck=False)

    def send_resetSystem(self, serviceTypeAck=True):
        helpers.OUT("resetSystem")
        return self.write_to_manager(packetType=0x15, payload=bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00"), serviceTypeAck=serviceTypeAck)

    def send_resetMote(self, macAddress, serviceTypeAck=True):
        helpers.OUT("resetMote")
        payload = bytearray(b"\x02") + macAddress
        return self.write_to_manager(packetType=0x15, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_subscribe(self, filter, unackFilter, serviceTypeAck=True):
        helpers.OUT("subscribe")
        payload = bytearray(struct.pack(">LL", filter, unackFilter))
        return self.write_to_manager(packetType=0x16, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getTime(self, serviceTypeAck=True):
        helpers.OUT("getTime")
        return self.write_to_manager(packetType=0x17, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_setNetworkConfig(self, networkID, apTxPower, frameProfile, maxMotes, baseBandwidth, downFrameMultVal, numParents, ccaMode, channelList, bbMode, bbSize, isRadioTest, bwMult, oneChannel, serviceTypeAck=False):
        helpers.OUT("setNetworkConfig")
        payload = bytearray(struct.pack(">HbBHHBBBHBBBHB", networkID, apTxPower, frameProfile, maxMotes, baseBandwidth, downFrameMultVal, numParents, ccaMode, channelList, bbMode, bbSize, isRadioTest, bwMult, oneChannel))
        return self.write_to_manager(packetType=0x1A, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_clearStatistics(self, serviceTypeAck=True):
        helpers.OUT("clearStatistics")
        return self.write_to_manager(packetType=0x1F, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_exchangeMoteJoinKey(self, macAddress, key, serviceTypeAck=True):
        helpers.OUT("exchangeMoteJoinKey")
        payload = macAddress + key
        return self.write_to_manager(packetType=0x21, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_exchangeNetworkId(self, id, serviceTypeAck=True):
        helpers.OUT("exchangeNetworkId")
        return self.write_to_manager(packetType=0x22, payload=bytearray([id]), serviceTypeAck=serviceTypeAck)

    def send_radiotestTx(self, testType, chanMask, repeatCnt, txPower, seqSize, sequenceDef, stationId, serviceTypeAck=True):
        helpers.OUT("radiotestTx")
        assert(len(sequenceDef) <= 10)
        payload = bytearray(struct.pack(">BHHbB", testType, chanMask, repeatCnt, txPower, seqSize))
        for seqDef in sequenceDef:
            payload += bytearray(struct.pack(">BH", seqDef.pkLen, seqDef.delay))
        payload += bytearray([stationId])
        return self.write_to_manager(packetType=0x23, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_radiotestRx(self, mask, duration, stationId, serviceTypeAck=True):
        helpers.OUT("radiotestRx")
        payload = bytearray(struct.pack(">HHB", mask, duration, stationId))
        return self.write_to_manager(packetType=0x25, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getRadiotestStatistics(self, serviceTypeAck=True):
        helpers.OUT("getRadiotestStatistics")
        return self.write_to_manager(packetType=0x26, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_setACLEntry(self, macAddress, joinKey, serviceTypeAck=True):
        helpers.OUT("setACLEntry")
        payload = bytearray(macAddress) + bytearray(joinKey)
        return self.write_to_manager(packetType=0x27, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getNextACLEntry(self, macAddress, serviceTypeAck=True):
        helpers.OUT("getNextACLEntry")
        payload = bytearray(macAddress)
        return self.write_to_manager(packetType=0x28, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_deleteACLEntry(self, macAddress, serviceTypeAck=True):
        helpers.OUT("deleteACLEntry")
        payload = bytearray(macAddress)
        return self.write_to_manager(packetType=0x29, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_pingMote(self, macAddress, serviceTypeAck=True):
        helpers.OUT("pingMote")
        payload = bytearray(macAddress)
        return self.write_to_manager(packetType=0x2A, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getLog(self, macAddress, serviceTypeAck=True):
        helpers.OUT("getLog")
        payload = bytearray(macAddress)
        return self.write_to_manager(packetType=0x2B, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_sendData(self, macAddress, priority, srcPort, dstPort, data, serviceTypeAck=True):
        helpers.OUT("sendData")
        payload = bytearray(macAddress) + bytearray(struct.pack(">BHHB", priority, srcPort, dstPort, 0)) + bytearray(data)
        return self.write_to_manager(packetType=0x2C, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getSystemInfo(self, serviceTypeAck=True):
        helpers.OUT("getSystemInfo")
        return self.write_to_manager(packetType=0x2E, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_getMoteConfig(self, macAddress, next, serviceTypeAck=True):
        helpers.OUT("Send getMoteConfig() to {}".format(helpers.u8s_to_str(macAddress, ":", "")))
        payload = bytearray(macAddress) + bytearray([next])
        return self.write_to_manager(packetType=0x2F, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getPathInfo(self, sourceMacAddress, destMacAddress, serviceTypeAck=True):
        helpers.OUT("getPathInfo")
        payload = bytearray(sourceMacAddress) + bytearray(destMacAddress)
        return self.write_to_manager(packetType=0x30, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getNextPathInfo(self, macAddress, filter, pathId, serviceTypeAck=True):
        helpers.OUT("getNextPathInfo")
        payload = bytearray(macAddress) + bytearray(struct.pack(">BH", filter, pathId))
        return self.write_to_manager(packetType=0x31, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_setAdvertising(self, activate, serviceTypeAck=True):
        helpers.OUT("setAdvertising")
        payload = bytearray([activate])
        return self.write_to_manager(packetType=0x32, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_setDownstreamFrameMode(self, frameMode, serviceTypeAck=True):
        helpers.OUT("setDownstreamFrameMode")
        payload = bytearray([frameMode])
        return self.write_to_manager(packetType=0x33, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getManagerStatistics(self, serviceTypeAck=True):
        helpers.OUT("getManagerStatistics")
        return self.write_to_manager(packetType=0x35, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_setTime(self, trigger, utcTime, serviceTypeAck=True):
        helpers.OUT("setTime")
        payload = bytearray([trigger]) + bytearray(utcTime)
        return self.write_to_manager(packetType=0x36, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_setCLIUser(self, role, password, serviceTypeAck=True):
        helpers.OUT("setCLIUser")
        payload = bytearray([role]) + bytearray(password)
        return self.write_to_manager(packetType=0x3A, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_sendIP(self, macAddress, priority, options, encryptedOffset, data, serviceTypeAck=True):
        helpers.OUT("sendIP")
        payload = bytearray(macAddress) + bytearray(struct.pack(">BBB", priority, options, encryptedOffset)) + bytearray(data)
        return self.write_to_manager(packetType=0x3B, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_restoreFactoryDefaults(self, serviceTypeAck=True):
        helpers.OUT("restoreFactoryDefaults")
        return self.write_to_manager(packetType=0x3D, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_getMoteInfo(self, macAddress, serviceTypeAck=True):
        helpers.OUT("getMoteInfo")
        payload = bytearray(macAddress)
        return self.write_to_manager(packetType=0x3E, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getNetworkConfig(self, serviceTypeAck=True):
        helpers.OUT("getNetworkConfig")
        return self.write_to_manager(packetType=0x3F, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_getNetworkInfo(self, serviceTypeAck=True):
        helpers.OUT("getNetoworkInfo")
        return self.write_to_manager(packetType=0x40, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_getMoteConfigById(self, moteId, serviceTypeAck=True):
        helpers.OUT("getMoteConfigById")
        payload = bytearray(struct.pack(">H", moteId))
        return self.write_to_manager(packetType=0x41, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_setCommonJoinKey(self, key, serviceTypeAck=False):
        helpers.OUT("setCommonJoinKey")
        payload = bytearray(key)
        return self.write_to_manager(packetType=0x42, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getIPConfig(self, serviceTypeAck=True):
        helpers.OUT("getIPConfig")
        return self.write_to_manager(packetType=0x43, payload=bytearray(), serviceTypeAck=serviceTypeAck)

    def send_setIPConfig(self, ipv6Address, mask, serviceTypeAck=True):
        helpers.OUT("setIPConfig")
        payload = bytearray(ipv6Address) + bytearray(mask)
        return self.write_to_manager(packetType=0x44, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_deleteMote(self, macAddress, serviceTypeAck=True):
        helpers.OUT("deleteMote")
        payload = bytearray(macAddress)
        return self.write_to_manager(packetType=0x45, payload=payload, serviceTypeAck=serviceTypeAck)

    def send_getMoteLinks(self, macAddress, idx, serviceTypeAck=True):
        helpers.OUT("getMoteLinks")
        payload = bytearray(macAddress) + bytearray(struct.pack(">H", idx))
        return self.write_to_manager(packetType=0x46, payload=payload, serviceTypeAck=serviceTypeAck)



# requirements
# google-cloud-pubsub==1.7.0
import json
import time
from google.auth import jwt
from google.cloud import pubsub_v1
from txp.common.configuration import SamplingWindow
from txp.common.edge import EdgeDescriptor, EdgeType
from txp.devices.drivers.mock.mock_driver import MockDriver, MockDriversHost
import base64
from multiprocessing.managers import BaseManager
from txp.devices.package_queue import PackageQueueProxy, PackageQueue
import multiprocessing
import datetime
from txp.common.configuration import SamplingParameters
from txp.devices.sampling_task import SamplingTask

# --- Base variables and auth path
CREDENTIALS_PATH = "../../../../common/credentials/pub_sub_to_bigquery_credentials.json"
PROJECT_ID = "tranxpert-mvp"
TOPIC_ID = "txp-telemetry-realtime-test"


# --- PubSub Utils Classes
class PubSubPublisher:
    def __init__(self, credentials_path, project_id, topic_id):
        credentials = jwt.Credentials.from_service_account_info(
            json.load(open(credentials_path)),
            audience="https://pubsub.googleapis.com/google.pubsub.v1.Publisher"
        )
        self.project_id = project_id
        self.topic_id = topic_id
        self.publisher = pubsub_v1.PublisherClient(credentials=credentials)
        self.topic_path = self.publisher.topic_path(self.project_id, self.topic_id)

    def publish(self, data: str, gateway_id):
        result = self.publisher.publish(self.topic_path, data, gatewayId=gateway_id)
        return result


# --- Main publishing script
def main(edge, sampling_params, host, q_proxy):
    publisher = PubSubPublisher(CREDENTIALS_PATH, PROJECT_ID, TOPIC_ID)

    task1 = SamplingTask("92", sampling_params, "tranxpert-mvp")
    host.receive_sampling_tasks(
        {
            edge.logical_id: task1
        }
    )
    host.start_sampling_windows()
    host.drivers[0].enable_observation_pulse()
    time.sleep(30)
    host.drivers[0].stop_sampling_windows()
    time.sleep(20)
    assert not q_proxy.is_empty()
    while not q_proxy.is_empty():
        package = q_proxy.get_package()
        proto = package.get_proto()
        encoded_payload = base64.b64encode(proto.SerializeToString())
        publisher.publish(encoded_payload, "test-mockup-gateway")
        time.sleep(30)


class CustomManager(BaseManager):
    """Multiprocessing Manager to register a PackageQueue as mock to the Gateway's queue."""
    pass


CustomManager.register("PackageQueue", PackageQueue, proxytype=PackageQueueProxy)


def host_discovered_identifiers():
    return [
        "001B638445E6",
        "001B638445E7",
        "001B638445E8"
    ]


def edge_descriptors():
    mock_1_descriptor: EdgeDescriptor = EdgeDescriptor(
        "mock_1",
        MockDriver.device_name(),
        str(EdgeType.STREAM_ONLY_DEVICE),
        {
            MockDriver._PHYSICAL_ID_PARAM_FIELD: "001B638445E6"
        },
        {
            "VibrationMock": {
                "dimensions": [1, 1024],
                "sampling_frequency": 100,
                "components_per_dimension": [
                    {
                        "components":
                        # 1 dimension, only 1 element in this list.
                            [
                                # Spectral components for this dimension.
                                {
                                    "amplitude": 1,
                                    "frequency": 7.0,
                                    "phase": 0,
                                    "component_type": "sine"
                                },
                                {
                                    "amplitude": 1,
                                    "frequency": 13.0,
                                    "phase": 0,
                                    "component_type": "sine"
                                }
                            ],
                    }
                ],
                "mode": 2
            },
        },
    )
    return [mock_1_descriptor]


def sampling_parameters() -> SamplingParameters:
    params = SamplingParameters(
        SamplingWindow(10, 5),
        datetime.datetime.now().date(),
        datetime.datetime.now().date(),
        "1111111",
        datetime.datetime.now().time(),
        (datetime.datetime.now() + datetime.timedelta(seconds=15)).time(),
        []
    )
    return params


def states_queue():
    """Returns a managed Queue to share acorss process.

    It mocks the states queue used by the DeviceManager to track
    the states of the DriversHost drivers.
    """
    queue = multiprocessing.Manager().Queue()
    return queue


def queue_proxy() -> PackageQueueProxy:
    manager = CustomManager()
    manager.start()
    q_proxy: PackageQueueProxy = manager.PackageQueue()
    return q_proxy


def mock_host(q_proxy, states_q) -> MockDriversHost:
    h = MockDriversHost(
        q_proxy, states_q,
        **{"discovered_identifiers": host_discovered_identifiers}
    )
    h.connect()
    return h


if __name__ == "__main__":
    q_proxy = queue_proxy()
    states_q = states_queue()
    host = mock_host(q_proxy, states_q)
    host._add_drivers(edge_descriptors())
    main(edge_descriptors()[0], sampling_parameters(), host, q_proxy)



from abc import ABC, abstractmethod
from txp.ml.prediction_service.state_managers.state_manager_output import RayStateManagerOutput
import json
from google.oauth2 import service_account
from txp.ml.prediction_service.state_managers.state_controller import StateController
import google.cloud.firestore as firestore
from txp.cloud import settings
from txp.ml.prediction_service.service_manager.actors_synchronizer import stateful_method
from txp.cloud import get_ml_event_log
from txp.common.ml.tasks import AssetState
import logging


class StateManager(ABC):
    """ An asset has multiple tasks all these tasks generate events, those events generates an state.
    This class in in charge of waiting for all the events to arrived and once all the events has arrived,
    this class creates an state.
    There is one state manager per each asset for a given tenant, this state manager classes are ad-hoc.
    It means someone has to code a new class for each new asset.

    Args:
        asset_id: machine id or asset id.
        events_and_states_dataset: BQ dataset name where events and state are going to be stored.
        backup_collection_name: Firestore collection name where backup is going to be stored.
        notifications_topic: PubSub topic id where notifications are going to be sent.
        credentials_str: GCP credentials string.
    """

    def __init__(self, asset_id, events_and_states_dataset, backup_collection_name, notifications_topic,
                 credentials_str):
        self.asset_id = asset_id
        json_dict_service_account = json.loads(credentials_str, strict=False)
        credentials = service_account.Credentials.from_service_account_info(json_dict_service_account)
        self.backup_collection_name = backup_collection_name
        self.firestore_db = firestore.Client(credentials=credentials, project=credentials.project_id)
        self.tasks = None
        self.events_cache = {}
        self.state_controller = StateController()
        self.load_backup()
        self.state_manager_output = RayStateManagerOutput.remote(events_and_states_dataset, notifications_topic,
                                                                 credentials_str)

    def backup(self):
        logging.info(f"Backing up state manager for {self.asset_id}")
        backup_collection = self.firestore_db.collection(self.backup_collection_name)
        state_controller_db_ref = backup_collection.document(settings.firestore.ray_state_controller_db_doc_name)
        state_controller_history_ref = backup_collection.document(
            settings.firestore.ray_state_controller_history_doc_name)
        state_manager_ref = backup_collection.document(settings.firestore.ray_state_manager_doc_name)
        if self.state_controller.states_db:
            logging.info(f"Backing up states db for {self.asset_id}")
            state_controller_db_ref.update({self.asset_id: self.state_controller.states_db})
        if self.state_controller.states_history:
            logging.info(f"Backing up states history for {self.asset_id}")
            state_controller_history_ref.update({self.asset_id: self.state_controller.states_history})
        if self.events_cache:
            logging.info(f"Backing up events cache for {self.asset_id}")
            state_manager_ref.update({self.asset_id: self.events_cache})
        logging.info(f"State manager backed up fo {self.asset_id}")

    def load_backup(self):
        logging.info(f"loading back up for {self.asset_id}")
        backup_collection = self.firestore_db.collection(self.backup_collection_name)
        state_controller_db_ref = backup_collection.document(settings.firestore.ray_state_controller_db_doc_name)
        state_controller_history_ref = backup_collection.document(
            settings.firestore.ray_state_controller_history_doc_name)
        state_manager_ref = backup_collection.document(settings.firestore.ray_state_manager_doc_name)

        state_controller_db = state_controller_db_ref.get({self.asset_id}).to_dict()
        if state_controller_db:
            self.state_controller.states_db = state_controller_db[self.asset_id]
            state_controller_db_ref.update({self.asset_id: firestore.DELETE_FIELD})
            logging.info(f"State controller db loaded for {self.asset_id}")
        else:
            logging.info(f"There is no previous state controller db entries for {self.asset_id}")

        state_controller_history = state_controller_history_ref.get({self.asset_id}).to_dict()
        if state_controller_history:
            self.state_controller.states_history = state_controller_history[self.asset_id]
            state_controller_history_ref.update({self.asset_id: firestore.DELETE_FIELD})
            logging.info(f"State controller history loaded for {self.asset_id}")
        else:
            logging.info(f"There is no previous state controller history entries for {self.asset_id}")

        state_manager = state_manager_ref.get({self.asset_id}).to_dict()
        if state_manager:
            self.events_cache = state_manager[self.asset_id]
            state_manager_ref.update({self.asset_id: firestore.DELETE_FIELD})
            logging.info(f"events cache loaded for {self.asset_id}")
        else:
            logging.info(f"There is no previous events cache for {self.asset_id}")
        logging.info(f"Backup loaded for {self.asset_id}")

    @abstractmethod
    def get_state(self, events) -> AssetState:
        pass

    def process_new_state(self, state, gateway_task_id):
        self.insert_state_to_bigquery(state)
        self.notify_new_state(state)

    def insert_events_to_bigquery(self, events):
        self.state_manager_output.insert_events_to_bigquery.remote(events)

    def insert_state_to_bigquery(self, state):
        self.state_manager_output.insert_state_to_bigquery.remote(state)

    def notify_new_state(self, state):
        self.state_manager_output.publish_to_pubsub.remote(state)

    def check_events_completeness(self, events):
        events_tasks = [event["task_id"] for event in events]
        for task in self.tasks:
            if task not in events_tasks:
                logging.info(f"There is no event for task_id: {task}, event is not completed")
                return False
        return True

    def store_event(self, event):
        observation_timestamp = str(event["observation_timestamp"])
        if observation_timestamp in self.events_cache:
            self.events_cache[observation_timestamp].append(event)
        else:
            self.events_cache[observation_timestamp] = [event]
        logging.info(f"Event {get_ml_event_log(event)}, stored at state manager cache")

    def delete_events_from_cache(self, observation_timestamps):
        events = []
        for observation_timestamp_ in observation_timestamps:
            observation_timestamp = str(observation_timestamp_)
            logging.info(f"Deleting events at state manager cache for {observation_timestamp}")
            events += self.events_cache[observation_timestamp]
            self.events_cache.pop(observation_timestamp)
        if events:
            self.insert_events_to_bigquery(events)

    @stateful_method
    def process_event(self, event, gateway_task_id, sampling_window_index, number_of_sampling_windows):
        logging.info(f"Trying to generate state for {get_ml_event_log(event)}")
        self.store_event(event)
        new_state = None
        observation_timestamp = str(event["observation_timestamp"])
        events = self.events_cache[observation_timestamp]
        logging.info(f"Checking event completeness for {get_ml_event_log(event)}")
        logging.info(f"{events}, were found")
        if self.check_events_completeness(events):
            logging.info(f"All necessary events were found when {get_ml_event_log(event)}, arrived generating a new state")
            generated_state = json.loads(self.get_state(events).json())
            logging.info(f"new state generated {generated_state}")

            self.state_controller.save_state(generated_state, gateway_task_id, sampling_window_index,
                                             number_of_sampling_windows)
            new_state = self.state_controller.get_state(gateway_task_id)
            processed_states = []
            while new_state is not None:
                logging.info(f"Processing new state {new_state}")
                self.process_new_state(new_state, gateway_task_id)
                logging.info(f"State processed {new_state}")
                self.state_controller.enqueue_state(new_state, gateway_task_id)
                logging.info(f"State enqueued at states history {new_state}")
                processed_states.append(new_state["observation_timestamp"])
                self.state_controller.clear_history(gateway_task_id)
                new_state = self.state_controller.get_state(gateway_task_id)
            self.delete_events_from_cache(processed_states)
        else:
            logging.info(f"There are still missing events {get_ml_event_log(event)}")
        return new_state

    def get_last_states(self, number_of_states, gateway_task_id):
        return self.state_controller.get_states_history(gateway_task_id, number_of_states)



import datetime
import matplotlib.pyplot as plt
import txp.cloud.controllers.bigquery as bigquery_controllers
import pytz
import logging
import streamlit as st
import plotly.graph_objects as go
import numpy as np
import os
import pandas as pd

plt.rcParams.update({"font.size": 6.5})
current_directory = os.getcwd()

# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


def mapping_data_from_data_parent(parent):
    list_of_element = []
    for element in parent:
        data = element.get().to_dict()
        list_of_element.append(data)
    return list_of_element


def change_utc_timezone(timestamp):
    utc = pytz.timezone("UTC")
    timezone = pytz.timezone("America/Mexico_City")
    date_time = pd.to_datetime(timestamp)
    localized_timestamp = utc.localize(date_time)
    new_timezone = localized_timestamp.astimezone(timezone)
    return new_timezone


def get_last_temperature_values(
    tenant_id, table_name, current_edge, current_perception, client,
):
    timezone = pytz.timezone("America/Mexico_City")

    last_hour = datetime.datetime.now() - datetime.timedelta(hours=1)
    localized_last_hour = timezone.localize(last_hour)

    now = timezone.localize(datetime.datetime.now())

    df_temperature_values = bigquery_controllers.get_all_records_within_interval(
        tenant_id,
        table_name,
        current_edge,
        current_perception,
        localized_last_hour.astimezone(pytz.utc),
        now.astimezone(pytz.utc),
        "observation_timestamp",
        client,
    )
    if not df_temperature_values.empty:
        temperature_values = {}
        for index, row in df_temperature_values.iterrows():
            timestamp = change_utc_timezone(row['package_timestamp'])
            for dimension in row["data"]:
                temperature_values[timestamp] = float(dimension["values"])
        return temperature_values
    else:
        logging.info(f"No hay data disponible")
        return None


def actual_machine_temperature(last_temperature_values):
    if not last_temperature_values is None:
        oldest_value = min(last_temperature_values.keys())
        youngest_value = max(last_temperature_values.keys())

        actual_and_last_temperature_difference = np.subtract(
            last_temperature_values[youngest_value],
            last_temperature_values[oldest_value],
        )
        st.metric(
            label="\n",
            value=f"{last_temperature_values[youngest_value]} °C",
            delta=f"{actual_and_last_temperature_difference} °C",
            delta_color="inverse",
        )
    else:
        st.info(f"Sin registro en la última hora.")


def plot_asset_temperatures(name, temperature_values, temperature_dates):
    if not temperature_values is None:
        strtimes = temperature_dates

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=strtimes[::-1],
                y=temperature_values[::-1],
                marker=dict(
                    size=16,
                    color=temperature_values[::-1],
                    colorscale="portland",
                    colorbar=dict(
                        title="Temperatura",
                        tickvals=[min(temperature_values), max(temperature_values)],
                        ticktext=["high", "low"],
                        ticks="outside",
                    ),
                ),
                mode="markers",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=strtimes[::-1],
                y=temperature_values[::-1],
                mode="lines",
                line=dict(color="grey", width=2, dash="dash"),
            )
        )

        fig.update_layout(margin=dict(l=0, r=0, b=0, t=0), showlegend=False)
        return fig

    else:
        return None



"""
This module contains the support for the DriversHost conceptualization.
"""

# ============================ imports ==========================================
import multiprocessing
import threading
from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Optional
from txp.devices.drivers.driver import DriverState, Driver, StateReport
from txp.common.edge import EdgeDescriptor
from txp.devices.drivers.icomox.icomox_driver import IcomoxDriver
from txp.devices.sampling_task import SamplingTask
from txp.devices.package_queue import PackageQueueProxy
from txp.common.edge import DeviceId
import queue
from collections import defaultdict

from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# DriversHost class
# ==============================================================================
class DriversHost(ABC):
    """A Drivers Host is a concept that represents a set of remote instances
    of Drivers.

    The 'Remote' word implies that the Drivers within the DriversHost are running
    on its own separated process, with the main responsibility of collecting samples
    from devices.

    The DriversHost can also encapsulate other required instances like managers
    or controllers that interact with the physical devices associated with the
    hosted edge_descriptors.
    """

    def __init__(self, gateway_packages_queue: PackageQueueProxy, drivers_state_queue: multiprocessing.Queue, error_timer_interval_seconds, **kwargs):
        """
        Args:
            gateway_packages_queue: The Gateway package queue used to collect the GatewayPackages.
            drivers_state_queue: The Gateway queue used to collect the driver states.
            **kwargs: named parameters specific for each DriversHost implementation.
        """
        self.gateway_package_queue: PackageQueueProxy = gateway_packages_queue
        self.drivers: List[Driver] = []
        self._drivers_index: Dict[str, Driver] = {}
        self.drivers_state_queue = drivers_state_queue
        """edge_descriptors: internal list of hosted edge_descriptors."""
        self.drivers_revering_from_error: List[str] = []
        self.error_timer_interval_seconds = error_timer_interval_seconds
        self._sampling_started = False
        log.debug(f"DrivesHost {self.get_driver_host_name()} running in pid: {multiprocessing.current_process().pid}")

        self._completed_driver_pulses_queue: threading.queue = queue.Queue()
        """_completed_driver_pulses_queue: internal queue for the host to receive the report 
        of completed clock pulses by the drivers in an sampling window. 
        Note: Currently the reports are the edges logical ids."""

        self._completed_clocks_notif_queue: Dict[str, int] = {}

    @abstractmethod
    def connect(self):
        """This method will connect the DriversHost with the physical device that
        is connected to the Gateway hardware unit.
        """
        pass

    @abstractmethod
    def disconnect(self):
        """This method will disconnect the DriversHost from the physical device that
        is connected to the Gateway hardware unit.
        """
        pass

    @abstractmethod
    def _add_drivers(self, edges: List[EdgeDescriptor]):
        """Method to be defined by the child classes in order to instantiate the drivers
        with logic-specific code."""
        pass

    def add_drivers(self, edges: List[EdgeDescriptor]):
        """Adds the set of drivers to the host for the specified edge descriptors,
        and creates the necessary internal tables.

        Args:
            edges: List of EdgeDescriptor for the edges.
        """
        self._add_drivers(edges)
        for driver in self.drivers:
            self._drivers_index.update({
                driver.logical_id: driver
            })
            self._completed_clocks_notif_queue[driver.logical_id] = 0

    @abstractmethod
    def discover_physical_ids(self) -> List[DeviceId]:
        """Discovers physical IDs connected available in the system for the Host
        type of device.

        Returns:
            A list of the discovered DeviceId in the physical layer.
        """
        pass

    @abstractmethod
    def get_driver_host_name(self):
        pass

    @abstractmethod
    def take_fix_error_action(self, serial):
        pass

    def _driver_errored_timeout_timer(self,serial):
        driver = list(filter(lambda driver: driver.logical_id == serial, self.drivers)).pop()
        state = driver.get_state()
        if state == DriverState.RECOVERING:
            driver.update_driver_state(DriverState.ERRORED)
            self._add_drivers_state_to_queue(StateReport(serial,DriverState.ERRORED))
        elif state == DriverState.CONNECTED:
            driver.start_sampling_windows()
            log.info("driver was able to be recovered to managed state {} by host serial {} resuming pending tasks".format(state, serial))
        self.drivers_revering_from_error.remove(serial)

    #TODO: try change signature and pass driver instance as a method variable so full control on driver on the error processing. instead of indexing
    def _on_driver_state_changed(self, serial, driver_state):
        if driver_state == DriverState.RECOVERING:
            if self.get_driver_host_name() == IcomoxDriver.device_name():
                self._process_recovery_driver_state(serial)
            else:
                log.warning("{} device does not support {}. it didnt get any signals during a Sampling Window".format(serial,driver_state))
        else:
            log.warning("{} device state has changed new state is {}".format(serial,driver_state))
            driver = list(filter(lambda driver: driver.logical_id == serial, self.drivers)).pop()
            driver.update_driver_state(driver_state)
            self._add_drivers_state_to_queue(StateReport(serial, driver_state))

    def _process_recovery_driver_state(self, serial):
        log.warning("driver on state error started for {}".format(serial))
        driver = list(filter(lambda driver: driver.logical_id == serial, self.drivers)).pop()
        # driver.stop_sampling_windows()
        driver.update_driver_state(DriverState.RECOVERING)
        self._add_drivers_state_to_queue(StateReport(serial, DriverState.RECOVERING))
        self.drivers_revering_from_error.append(serial)
        self.take_fix_error_action(serial)
        timer = threading.Timer(self.error_timer_interval_seconds, self._driver_errored_timeout_timer, args=(serial,))
        timer.start()

    def _add_drivers_state_to_queue(self, state_report: StateReport):
        self.drivers_state_queue.put(state_report)

    def receive_sampling_tasks(self, tasks: Dict[str, SamplingTask]):
        """Receives a list of sampling tasks to be executed by the
        drivers in this hosts.

        Args:
            tasks: Dictionary with the drivers serial as key and the
            corresponding sampling tasks as value.
        """
        for driver in self.drivers:
            if driver.driver_state == DriverState.CONNECTED or driver.driver_state == DriverState.SAMPLING:
                if driver.logical_id in tasks:
                    task = tasks[driver.logical_id]
                    log.info("Sending task to {}".format(
                        driver.logical_id,
                        task.sampling_window.observation_time,
                        task.sampling_window.sampling_time)
                    )
                    driver.receive_sampling_task(task)
            else:
                log.warning("Driver Serial {} is not ready to receive tasks state is {}".format(driver.logical_id, driver.driver_state))

    def receive_sampling_tasks_list(self, tasks: List[Tuple]):
        """Receive a list of sampling tasks to be executed by the drivers on this host."""
        for logical_id, sampling_task in tasks:
            if logical_id in self._drivers_index:
                driver = self._drivers_index[logical_id]
                if driver.driver_state == DriverState.CONNECTED or driver.driver_state == DriverState.SAMPLING:
                    log.info("Sending task to {}, with observation {}s and sampling {}s".format(
                        driver.logical_id,
                        sampling_task.sampling_window.observation_time,
                        sampling_task.sampling_window.sampling_time)
                    )
                    driver.receive_sampling_task(sampling_task)
                else:
                    log.warning("Driver Serial {} is not ready to receive tasks state is {}".format(driver.logical_id,
                                                                                                    driver.driver_state))
            else:
                log.error(f"Unexpected task received in host {self.get_driver_host_name()} for driver {logical_id}. "
                          f"The Driver doesn't belong to this host.")

    def connect_driver(self, edge_serial: str):
        """Connects the driver with the specified serial"""
        driver: Driver = next(
            (driver for driver in self.drivers if driver.logical_id == edge_serial), None
        )
        if driver is None:
            log.warning(
                f"Trying to connect driver {edge_serial}, but the driver was not found in the host."
            )
            return

        driver.disconnect_device()

    def connect_drivers(self):
        """Connects all the edge_descriptors found in the hosts."""
        for driver in self.drivers:
            driver.connect_device()

    def disconnect_driver(self, edge_serial: str):
        """Disconnect the driver with the specified serial"""
        driver: Driver = next(
            (driver for driver in self.drivers if driver.logical_id == edge_serial), None
        )
        if driver is None:
            log.warning(
                f"Trying to disconnect driver {edge_serial}, but the driver was not found in the host."
            )
            return

    def disconnect_drivers(self):
        """Disconnects all the edge_descriptors found in the hosts."""
        for driver in self.drivers:
            driver.disconnect_device()

    def get_drivers_state(self) -> List[DriverState]:
        """Returns the states for the edge_descriptors of this hosts.

        Returns:
            The List of driver states.
        """
        states = [driver.get_state() for driver in self.drivers]
        return states

    def get_drivers_descriptors(self) -> List[EdgeDescriptor]:
        """Returns the description of the hosted edge_descriptors."""
        descriptors = [driver.get_descriptor() for driver in self.drivers]
        return descriptors

    def start_sampling_windows(self, logical_id: Optional[str] = None):
        """Starts the sampling windows for the hosted drivers.

        Args:
            logical_id: optional unique logical_id. If provided, only that
                driver will be enabled.
        """
        self._sampling_started = True  # At least one driver is sampling.

        if not logical_id:
            for driver in self.drivers:
                if driver.driver_state == DriverState.CONNECTED:
                    driver.start_sampling_windows()
        else:
            self._drivers_index[logical_id].start_sampling_windows()

    def stop_sampling_windows(self):
        """Starts the sampling windows for the hosted drivers"""
        self._sampling_started = False
        for driver in self.drivers:
            driver.stop_sampling_windows()

    def get_driver_observation_pulse(self, logical_id: str) -> int:
        """Returns the observation pulse of the driver."""
        if logical_id not in self._drivers_index:
            log.error(f"Host {self.get_driver_host_name()} was requested to obtain "
                      f"pulse from driver {logical_id}, but the Driver was not found")
            return 0
        pulse = self._drivers_index[logical_id].observation_pulse()
        return pulse

    def notify_sampling_window_completed(self, logical_id: str) -> None:
        """Notifies to the driver that the sampling window was completed"""
        if logical_id not in self._drivers_index:
            log.error(f"Host {self.get_driver_host_name()} was requested to notify sampling window completion "
                      f"to driver {logical_id}, but the Driver was not found")
        else:
            self._drivers_index[logical_id].notify_sampling_window_completed()

    def enable_driver_clock_pulse(self, logical_id: str) -> None:
        """Enables the drivers internal next clock pulse in order to authorize a synchronized
        and controlled driver observation."""
        if logical_id not in self._drivers_index:
            log.error(f"Host {self.get_driver_host_name()} was requested to enable "
                      f"pulse from driver {logical_id}, but the Driver was not found")
        else:
            self._drivers_index[logical_id].enable_observation_pulse()

    def disable_current_pulse(self, logical_id: str) -> None:
        """Enable the signal to stop the current sampling collect."""
        if logical_id not in self._drivers_index:
            log.error(f"Host {self.get_driver_host_name()} was requested to stop "
                      f"pulse from driver {logical_id}, but the Driver was not found")
        else:
            self._drivers_index[logical_id].disable_current_pulse()

    def is_virtual_driver(self, logical_id: str) -> bool:
        """Returns true if the driver with logical id is virtual."""
        if logical_id not in self._drivers_index:
            log.error(f"Host {self.get_driver_host_name()} was requested to check driver type "
                      f"from driver {logical_id}, but the Driver was not found")
            return False
        return self._drivers_index[logical_id].is_virtual()

    def is_actuator_device(self, logical_id: str) -> bool:
        """Returns true if the logical_id belongs to an actuator device.

        Note: The only Actuator device currently supported is the general
        descriptor for a Robot Driver."""
        return False

    def get_driver_completed_pulses(self, logical_id: str) -> int:
        """Returns the last packages report for the provided logical_id.
        """

        # Update the internal count first
        if self._completed_driver_pulses_queue.qsize():
            log.info(f"Drivers Host {self.get_driver_host_name()} has reports to process.")
            while self._completed_driver_pulses_queue.qsize():
                edge_logical_id = self._completed_driver_pulses_queue.get()
                self._completed_clocks_notif_queue[edge_logical_id] += 1
                log.info(f"Registering new package count for {edge_logical_id}:"
                         f" {self._completed_clocks_notif_queue[edge_logical_id]}")

        packages_count = self._completed_clocks_notif_queue[logical_id]
        return packages_count

    def clear_driver_registered_packages(self, logical_id: str):
        self._completed_clocks_notif_queue[logical_id] = 0



"""
This module defines the helper function to draw the devices pairing
panel in the browser.
"""
# =====================imports=============================================
import streamlit as st
import time
from txp.devices.ux.remote_gateway_client import remote_gateway_client
from txp.devices.gateway_states_enums import *
import txp.devices.ux.ux_states_protocol as ux_states
from txp.common.edge.common import EdgeDescriptor
import matplotlib.pyplot as plt
from typing import List, Set, Tuple
import cv2
import numpy as np
import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


# ==================== constant data ==========================================
_machines_ids: List[str] = []
_captured_virtual_pos: Set[str] = set()

# ==================== Helper functions ==========================================
def _get_captured_image(image_bytes):
    """Returns the matplotlib figure to render an image that is in memory a bytes"""
    image = np.frombuffer(image_bytes, dtype=np.uint8)
    image = cv2.imdecode(image, flags=1)
    image = image[:, :, ::-1]
    return image


# ==================== Functions related to NON virtual pairing ====================
def _pair_normal_device_button_clicked(edge_logical_id: str):
    new_state = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.PairingSelectedDevice
    new_state.props['device_to_pair'] = edge_logical_id
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"Sending device to pair information to Gateway."
              f" Device ID: {edge_logical_id}")


def _confirm_normal_pairing_button_clicked():
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.PairingConfirmed
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"Sending device pairing confirmation to Gateway."
              f" Device ID: {new_state.props['device_to_pair']}")


def _cancel_normal_pairing_button_clicked():
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.PairingCancelled
    new_state.props['device_to_pair'] = None
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"Sending device pairing cancellation to Gateway.")


def _retry_normal_button_clicked():
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.SelectDeviceToPair
    new_state.props['device_to_pair'] = None
    st.session_state.gateway_state = new_state
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"Retry button was clicked for current device.")


# ==================== Functions related to VIRTUAL pairing ====================
def _virtual_camera_verification_clicked():
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualCameraConnectionConfirmed
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User confirmed virtual camera connection."
              f" Sending information to Gateway")


def _virtual_thermocamera_verification_clicked():
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualThermocameraConnectionConfirmed
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User confirmed virtual thermocamera connection."
              f" Sending information to Gateway")


def _enable_torque_button_clicked():
    start_time = time.time()
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualEnableTorque
    end_time = time.time()
    log.info(f"Elapsed Time copying state in UX process: {end_time - start_time}s")
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User enabled torque. Sending information to Gateway")


def _disable_torque_button_clicked():
    start_time = time.time()
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualDisableTorque
    end_time = time.time()
    log.info(f"Elapsed Time copying state in UX process: {end_time-start_time}s")
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User disabled torque. Sending information to Gateway")


def _capture_position_button_clicked():
    start_time = time.time()
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualCapturePosition
    end_time = time.time()
    log.info(f"Elapsed Time copying state in UX process: {end_time-start_time}s")
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User selected to capture position. Sending information to Gateway")


def _discard_position_button_clicked():
    start_time = time.time()
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualRejectCapturedPosition
    end_time = time.time()
    log.info(f"Elapsed Time copying state in UX process: {end_time - start_time}s")
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User selected to discard position. Sending information to Gateway")


def _confirm_position_button_clicked(position: Tuple):
    start_time = time.time()
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualConfirmCapturedPosition
    end_time = time.time()
    machine = st.session_state.virtual_edge_selected_asset
    new_state.props["virtual_edge_asset"] = machine
    log.info(f"Elapsed Time copying state in UX process: {end_time - start_time}s")
    _captured_virtual_pos.add(EdgeDescriptor.get_hashed_virtual_id(position))
    st.session_state.gateway_state.props['virtual_production'].position_captured = tuple()
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User confirmed captured position. Sending information to Gateway.")


def _finish_virtual_pairing_button_clicked():
    new_state: ux_states.UxDevicePairingState = st.session_state.gateway_state.copy()
    new_state.status = DevicePairingStatus.VirtualConfirmAllPositions
    _captured_virtual_pos.clear()
    remote_gateway_client.send_state_to_gateway(new_state)
    log.debug(f"User confirmed ALL virtual position. Sending information to Gateway.")


# ==================== Rendering of devices pairing panel ====================
def render_devices_pairing():
    """The Devices Pairing panel is used to show interactive widgets
    when a pairing is required, that means when the Gateway State
    is Devices Pairing.
    """
    st.markdown("-------------")
    st.subheader("Emparejar Dispositivos")

    if not _machines_ids:
        for machine in st.session_state.gateway_state.machines:
            _machines_ids.append(machine.machine_id)

    # If Device Pairing is the Current State, then the widgets are shown.
    # The State should keep a track on how many devices have been paired,
    # to decide when all the devices are paired.
    if st.session_state.gateway_state.state == GatewayStates.DevicesPairing:
        paired_devices = st.session_state.gateway_state.props['paired_devices']
        expected_devices = st.session_state.gateway_state.props['expected_devices']
        st.markdown(
            f"**Dispositivos configurados**: {len(expected_devices)}"
        )
        st.markdown(
            f"**Dispositivos emparejados**: {len(paired_devices)}/{len(expected_devices)}"
        )

        # There are devices to pair
        if (
            len(paired_devices) < len(expected_devices)
        ):
            unpaired_devices = expected_devices - paired_devices  # set difference

            device = st.selectbox(
                label="Seleccionar dispoositivo para emparejar",
                key="pair_device_select_box",
                options=list(unpaired_devices),
            )

            # Status: User will select a device to pair
            if st.session_state.gateway_state.status == DevicePairingStatus.SelectDeviceToPair:
                pair_button = st.button(
                    label="Emparejar", key="pair_device_button",
                    on_click=_pair_normal_device_button_clicked, args=(device,)
                )

            # Status: User selected to pair a non-virtual device, and it's being paired by the
            # Gateway
            if st.session_state.gateway_state.status == DevicePairingStatus.PairingSelectedDevice:
                my_bar = st.progress(0)
                for percent_complete in range(10):
                    time.sleep(1)
                    my_bar.progress(percent_complete + 1)

            # The Gateway found successfully a non-virtual device, and request confirmation.
            if st.session_state.gateway_state.status == DevicePairingStatus.RequestConfirm:
                st.success(f"Se encontró un dispositivo conectado para el serial "
                           f"{st.session_state.gateway_state.props['device_to_pair']}")
                c1, c2 = st.columns([.2, 1])  # trick: https://discuss.streamlit.io/t/two-buttons-on-the-same-line/2749/4
                with c1:
                    st.button(
                        label="Confirmar",
                        key="confirm_pairing_button",
                        on_click=_confirm_normal_pairing_button_clicked,
                    )
                with c2:
                    st.button(
                        label="Cancelar",
                        key="cancel_pairing_button",
                        on_click=_cancel_normal_pairing_button_clicked,
                    )

            # The Gateway found an error pairing a non-virtual device
            if st.session_state.gateway_state.status == DevicePairingStatus.PairingError:
                st.error(f"Se encontró un error mientras se emparejaba el dispositivo. "
                         f"Verifique que solo hay 1 dispositivo conectado e intente nuevamente")
                st.button(label="Reintentar", on_click=_retry_normal_button_clicked)

            # ======================== Virtual pairing rendering =======================================
            # The Gateway request to the user the connection of the normal camera
            if st.session_state.gateway_state.status == DevicePairingStatus.RequestVirtualCameraConnection:
                st.markdown("El dispositivo es un dispositivo virtual."
                            " Empezará la producción del dispositivo virtual")
                st.info("Por favor conecte la camara fotográfica al puerto usb")
                verify_camera_btn = st.button(
                    "Verificar Camara",
                    key="virtual_camera_verification",
                    help="Presione este botón cuando haya conectado la cámara",
                    on_click=_virtual_camera_verification_clicked,
                )
                if verify_camera_btn:  # The button was pressed. The Gateway is trying to find the camera.
                    my_bar = st.progress(0)
                    for percent_complete in range(10):
                        time.sleep(1)
                        my_bar.progress(percent_complete + 1)

            # The Gateway successfully connected the normal camera for the virtual edge.
            if st.session_state.gateway_state.status == DevicePairingStatus.VirtualCameraProducedSuccessfully:
                st.success("La camara fotografica fue encontrada con éxito para el dispositivo robot")
                st.markdown("Espere mientras el Gateway procesa la información")

            # There was an error connecting the normal camera for the virtual edge.
            if st.session_state.gateway_state.status == DevicePairingStatus.VirtualCameraConnectionError:
                st.warning("Hubo un error detectando la cámara fotográfica")
                st.text("Espere mientras el Gateway procesa la información")

            # The Gateway now requested to the user the connection of Thermocamera
            if st.session_state.gateway_state.status == DevicePairingStatus.RequestVirtualThermocameraConnection:
                st.markdown("Por favor conecte la cámara termográfica al puerto usb")
                verify_thermocamera_btn = st.button(
                    "Verificar Termocamara", key="virtual_thermocamera_verification",
                    help="Presiones este botón cuando haya conectado la cámara termográfica",
                    on_click=_virtual_thermocamera_verification_clicked
                )
                if verify_thermocamera_btn:  # The button was pressed. The Gateway is trying to find the camera.
                    my_bar = st.progress(0)
                    for percent_complete in range(10):
                        time.sleep(1)
                        my_bar.progress(percent_complete + 1)

            # The thermographic camera was connected successfully
            if st.session_state.gateway_state.status == DevicePairingStatus.VirtualThermocameraProducedSuccessfully:
                st.success("La camara termográfica fue encontrada con éxito para el dispositivo robot")
                st.markdown("Espere mientras el Gateway procesa la información")

            if st.session_state.gateway_state.status == DevicePairingStatus.VirtualThermocameraConnectionError:
                st.warning("Hubo un error detectando la cámara termográfica")
                st.text("Espere mientras el Gateway procesa la información")

            # The Gateway now request the user for the next action in the virtual production
            # This is a "loop" of interactions with the user, until the user has produced all
            # the virtual edges.
            if st.session_state.gateway_state.status == DevicePairingStatus.VirtualRequestNextStepFromUser:
                col1, col2 = st.columns(2)
                production_state = st.session_state.gateway_state.props['virtual_production']
                is_torque_enabled = production_state.torque_enabled
                position_captured = bool(production_state.position_captured)
                at_least_one = bool(production_state.virtual_positions_produced)
                invalid_pos = False
                with col1:
                    st.text("Aquí se mostrará la imagen capturada")
                    if position_captured:
                        st.markdown(f"**Posición capturada**: {production_state.position_captured}")

                        if None in production_state.position_captured:  # The captured position has at least 1 None
                            st.warning("No es posible registrar la posición porque hay "
                                       "valores fuera de rango en los motores.")
                            invalid_pos = True

                        if EdgeDescriptor.get_hashed_virtual_id(production_state.position_captured) in _captured_virtual_pos:
                            st.warning("La posición capturada ya existe.")
                            invalid_pos = True

                        log.info("Rendering captured images in UI")
                        start_time = time.time()
                        if production_state.image_captured:
                            st.image(_get_captured_image(production_state.image_captured),
                                     caption='Imagen normal', output_format="JPEG")
                        else:
                            st.write("No se pudo captura imagen fotografica")

                        if production_state.thermographic_img_captured:
                            st.image(_get_captured_image(production_state.thermographic_img_captured),
                                     caption='Imagen Termográfica', output_format="JPEG")
                        else:
                            st.write("No se pudo capturar imagen termográfica")
                        end_time = time.time()
                        log.info(f"Total time spent rendering images: {end_time-start_time}")
                        st.selectbox(
                            "Asset asociado a la posición",
                            options=_machines_ids,
                            key="virtual_edge_selected_asset",
                        )

                    st.markdown("------")
                    st.markdown("**Posiciones Capturadas:**")
                    # TODO Format output.
                    st.markdown(f"""{list(map(
                        lambda entry: f'{entry[0].get_virtual_edge_information()}, Machine {entry[1]}',
                        production_state.virtual_positions_produced
                    ))}""")

                with col2:
                    st.button('Activar Torque', on_click=_enable_torque_button_clicked, disabled=is_torque_enabled)
                    st.button('Desactivar Torque', on_click=_disable_torque_button_clicked, disabled=not is_torque_enabled)
                    st.button('Capturar Posición', on_click=_capture_position_button_clicked, disabled=position_captured)
                    st.button('Descartar Posición', on_click=_discard_position_button_clicked, disabled=not position_captured)
                    st.button('Guardar Posición', on_click=_confirm_position_button_clicked,
                              args=(production_state.position_captured,),
                              disabled=not position_captured and invalid_pos)
                    st.markdown("-------")
                    st.button("Terminar", on_click=_finish_virtual_pairing_button_clicked, disabled=not at_least_one)

    else:
        # In this case pairing was completed in the past, and shows the feedback from those operations

        st.markdown(
            f"**Dispositivos configurados**: {len(st.session_state.gateway_state.edges)}"
        )
        st.markdown(
            f"**Dispositivos emparejados**: {len(st.session_state.gateway_state.edges)}/"
            f"{len(st.session_state.gateway_state.edges)}"
        )



'''
Created on Mar 6, 2012

\author alushin
'''

class ConnectionError(Exception) :
    '''
    \brief Exception class associated with connection to the device.
    '''
    
    def __init__(self, reason):
        self.value = reason

    def __str__(self) :
        return "Connection Error: {0}".format(self.value)
        
class QueueError(Exception) :
    def __init__(self):
        pass

class CommandTimeoutError(Exception) :
    '''
    \brief Device timeout error (disconnection)
    '''
    def __init__(self, cmd):
        self.cmd = cmd
    def __str__(self) :
        return "Device timeout error for command {0}".format(self.cmd)
    
class APIError(Exception) :
    '''
    \brief Exception class associated with connection the API.
    '''
    
    def __init__(self, cmd, rc, desc=None):
        self.cmd   = cmd
        self.rc    = rc
        self.desc  = desc

    def __str__(self) :
        return "Command {0} returns RC={1} {2}".format(self.cmd,
                                                       self.rc,
                                                       self.desc)

class NotificationError(Exception) :
    def __init__(self, notifIds, params):
        self.notifIds = notifIds
        self.params  = params

    def __str__(self) :
        return "Notification {0} params {1}".format(self.notifIds, self.params)
    
class CommandError(Exception):
    '''
    \brief Exception class associated with an API definition.
    '''
    
    INVALID_COMMAND          = 1
    INVALID_SUBCOMMAND       = 2
    ONLY_SUBCOMMANDS         = 3
    NO_SUBCOMMANDS           = 4
    UNKNOWN_SUBCOMMAND       = 5
    NO_REQUEST               = 6
    NO_RESPONSE              = 7
    TOO_FEW_FIELDS           = 8
    TOO_MANY_FIELDS          = 9
    UNKNOWN_FIELD            = 10
    MALFORMED_FIELD          = 11
    TOO_FEW_BYTES            = 12
    VALUE_NOT_IN_OPTIONS     = 14
    
    descriptions = { 
        INVALID_COMMAND:     'Command does not exist', 
        INVALID_SUBCOMMAND:  'Subcommand does not exist', 
        ONLY_SUBCOMMANDS:    'This function has subcommands only', 
        NO_SUBCOMMANDS:      'This function does not have any subcommands', 
        UNKNOWN_SUBCOMMAND:  'This subcommand does not exist', 
        NO_REQUEST:          'No request format defined for this command', 
        NO_RESPONSE:         'No response format defined for this command', 
        TOO_FEW_FIELDS:      'Too few fields passed', 
        TOO_MANY_FIELDS:     'Too many fields passed', 
        UNKNOWN_FIELD:       'Unknown field', 
        MALFORMED_FIELD:     'Field malformed',
        TOO_FEW_BYTES:       'Too few bytes in received packet',
        VALUE_NOT_IN_OPTIONS:'The value of this field is not in the valid options',
    }
    
    def __init__(self,errorCode,details=None):
        self.errorCode  = errorCode
        self.details    = details
    
    def __str__(self):
        try:
            output = self.descriptions[self.errorCode]
            if self.details:
                output += ': ' + str(self.details)
            return output
        except KeyError:
            return "Unknown error: #" + str(self.errorCode)


"""
This module defines the classes used to exchange information between
the Gateway process and the UI process.
"""
# ==================== Imports ===================================
from dataclasses import dataclass
from typing import List, Set, Optional, Dict
from txp.common.edge import MachineMetadata, EdgeDescriptor
from txp.devices.gateway_states_enums import *
from txp.devices.drivers.driver import DriverState
import copy
import enum


# =================== Ux States Classes ====================================
# UX States classes are designed to represent the Gateway State in the UX process,
# along with attributes used to track the Ux state itself.
@dataclass
class UxDetails:
    class Kind(enum.Enum):
        Warning=1
        Info=2
        Error=3

    """A dataclass to hold and give structure to string messages that
    provide human readable details about some state
    
    This enumeration can be used with streamlit to provide visual feedback
    using st.warning, st.error or st.info.
    """
    kind: Kind
    message: str


@dataclass
class UxStateBase:
    """A base class which defines the common attributes for all the
    messages passed between the Gateway and the UX process"""
    state: GatewayStates
    """state: A member of the GatewayStates enumeration"""
    status: "Status"
    """status: A member of some Status enumeration defined in the 
        gateway enumerations"""
    details: Optional[UxDetails]
    """An optional UxDetails field to provide human readable details 
        about the current state"""


@dataclass
class UxStartupState(UxStateBase):
    """The UX state to work with the Gateway Startup state.

    In Startup, the UX will require from the user all the pre-conditioned
    values that are required to operate.

    All the values requested to the user will be individual attributes in
    this class.
    """
    gateway_id: str
    """The gateway device ID / Serial configured in the Cloud IoT management"""
    registry_id: str
    """The IoT registry to which the gateway belongs"""
    private_key_received: bool
    """private_key_received: True if the UX already received the private key file upload"""
    gateway_serial_received: bool
    """gateway_serial_received: True if the UX already received the Gateway Serial"""
    registry_name_received: bool
    """registry_name_received: True if the UX already received the IoT registry name"""
    project_id_received: bool
    project_id: str


@dataclass
class UxConfigurationReceptionState(UxStateBase):
    """Contains the received configurations from the cloud

    The Gateway will be updating the state as it
    receives the remote configurations.

    At this point, the Gateway will send the GatewayConfig
    instead of inividual attributes for the gateway device ID
    and the registry ID.
    """
    gateway_conf: "GatewayConfig"
    """gateway_conf: The GatewayConfig instance received from cloud."""
    machines: List[MachineMetadata]
    """machines: The Machines logical information received from cloud."""
    edges: List[EdgeDescriptor]
    """edges: The EdgeDescriptors received from cloud"""
    all_received: bool
    """all_received: True if all the configurations were received."""


@dataclass
class UxConfigurationValidationState(UxStateBase):
    """Contains the information used during the Configuration Validation Step.

    Note: Currently, the Gateway waits for the acknowledge of the user in order to move
    to the next state.
    This State allows Restart of the Gateway.
    """
    gateway_conf: "GatewayConfig"
    """gateway_conf: The GatewayConfig instance received from cloud."""
    machines: List[MachineMetadata]
    """machines: The Machines logical information received from cloud."""
    edges: List[EdgeDescriptor]
    """edges: The EdgeDescriptors received from cloud"""


@dataclass
class UxDevicePairingState(UxStateBase):
    gateway_conf: "GatewayConfig"
    machines: List[MachineMetadata]
    edges: List[EdgeDescriptor]
    props: Dict

    def copy(self) -> "UxDevicePairingState":
        return copy.deepcopy(self)

@dataclass
class UxPreparingDevicesState(UxStateBase):
    """Preparing Devices State to be received in the UI process
    when the gateway is preparing the devices.
    """
    gateway_conf: "GatewayConfig"
    machines: List[MachineMetadata]
    edges: List[EdgeDescriptor]
    devices_connection_table: Dict[str, DriverState]


@dataclass
class UxErrorState(UxStateBase):
    """The UxErrorState to communicate an error state to the UI process.
    """
    pass


@dataclass
class UxReadyState(UxStateBase):
    gateway_conf: "GatewayConfig"
    machines: List[MachineMetadata]
    edges: List[EdgeDescriptor]
    collected_packages: int
    sent_packages: int
    completed_package_cycles: int
    packages_received_by_edge: Dict[str, int]
    devices_connection_table: Dict[str, DriverState]


@dataclass
class UxErrorState(UxStateBase):
    """The UxErrorState to communicate an error state to the UI process.
    """
    pass


# =================== Ux Deprecated Mock Classes ====================================
# Deprecated classes used in the first version of the streamlit app to test the Raspberry OS.
# If some part of the UX is still mocked, it's using these classes.
# TODO. Remove this classes and create the final UXStateBase implementations for the rest of states.

@dataclass
class GatewayCommonMetadata:
    """A base class which shares attributes common for all the UX Gateway
    states classes.

    Each state will contain all this information.

    Note: if there's a piece of information that requires to be shared
    progressively across all states, it probably should be defined as an
    attribute here.

    TODO: Deprecated. This was used for mocking in the Raspberry image.
    """

    device_id: str
    """The gateway device ID"""
    registry_id: str
    """The IoT registry to which the gateway belongs"""
    state: GatewayStates
    """The gateway state value"""
    status: enum.Enum
    """The current status for the state"""

@dataclass
class UxSamplingState(GatewayCommonMetadata):
    """Contains information about the collected Packages, while being
    in the Sampling State.

    TODO: Information about packaging is better to have it as a GlobalState
        value. A State that the Gateway object uses to handle global
        information collected during the Gateway lifecycle.
        A reference to that State might be received here. That'll simplify a lot
        the states declared in the Ux.
    """

    gateway_conf: "GatewayConfig"
    machines: List[MachineMetadata]
    edges: List[EdgeDescriptor]
    paired_devices: Set[str]
    expected_devices: Set[str]
    collected_packages: int
    sent_packages: int
    completed_package_cycles: int
    packages_received_by_edge: Dict[str, int]



"""
This module defines the main function to be called from the CLI program
for the command: run_vicomox_test
"""
# ============================ imports =========================================
import queue

from txp.devices.drivers.driver import DriverState
from txp.devices.drivers.icomox.icomox_driver import IcomoxDriver
from txp.devices.drivers.icomox.icomox_driver_host import IcomoxDriversHost
from txp.common.edge import EdgeDescriptor, EdgeType, MachineMetadata
from txp.devices.sampling_task import SamplingTask
from txp.common.configuration import SamplingWindow
from txp.devices.package_queue import PackageQueue, PackageQueueProxy
import multiprocessing
from typing import Tuple
import sys
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Server process to MOCK the gateway's package queue.
# ==============================================================================
class ProcessesManager(multiprocessing.managers.BaseManager):
    """This class will be used to create the Driver, the SmartMeshManager and the
    PackageQueue in separated processes"""

    pass


ProcessesManager.register("PackageQueue", PackageQueue, proxytype=PackageQueueProxy)

# ==============================================================================
# Declares the EdgeDescriptors for the Icomox instance
# ==============================================================================
machine_metadata: MachineMetadata = MachineMetadata(
    "licuadora_1", "Clasica 4655", "Osterizer", ["icomox_1"]
)

icomox_1_descriptor: EdgeDescriptor = EdgeDescriptor(
    "icomox_1",
    EdgeType.STREAM_ONLY_DEVICE,
    IcomoxDriver.device_name(),
    {
        IcomoxDriver._PHYSICAL_ID_PARAM_FIELD : "12A79E50FF0E7D384128040038343038",
    },
    {
        "VibrationAcceleration": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 1024],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 3611.11
        },
        "VibrationSpeed": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 399.60
        },
        "Magnetometer": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 334.36
        },
        "Microphone": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [1, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 20312.5
        }
    }
)

icomox_2_descriptor: EdgeDescriptor = EdgeDescriptor(
    "icomox_2",
    EdgeType.STREAM_ONLY_DEVICE,
    IcomoxDriver.device_name(),
    {
        IcomoxDriver._PHYSICAL_ID_PARAM_FIELD: "12A79EA8370F5D383C23040038343038",
    },
    {
        "VibrationAcceleration": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 1024],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 3611.11
        },
        "VibrationSpeed": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 399.60
        },
        "Magnetometer": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 334.36
        },
        "Microphone": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [1, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 20312.5
        }
    }
)

icomox_3_descriptor: EdgeDescriptor = EdgeDescriptor(
    "icomox_3",
    EdgeType.STREAM_ONLY_DEVICE,
    IcomoxDriver.device_name(),
    {
        IcomoxDriver._PHYSICAL_ID_PARAM_FIELD: "12A79E3C872FFD382BEF040038343038",
    },
    {
        "VibrationAcceleration": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 1024],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 3611.11
        },
        "VibrationSpeed": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 399.60
        },
        "Magnetometer": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 334.36
        },
        "Microphone": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [1, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 20312.5
        }
    }
)

icomox_4_descriptor: EdgeDescriptor = EdgeDescriptor(
    "icomox_4",
    EdgeType.STREAM_ONLY_DEVICE,
    IcomoxDriver.device_name(),
    {
        IcomoxDriver._PHYSICAL_ID_PARAM_FIELD: "12A79E7CC70FDD382C1C040038343038",
    },
    {
        "VibrationAcceleration": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 1024],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 3611.11
        },
        "VibrationSpeed": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 399.60
        },
        "Magnetometer": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [3, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 334.36
        },
        "Microphone": {
            IcomoxDriver.PERCEPTION_DIMENSIONS_FIELD: [1, 512],
            IcomoxDriver.PERCEPTION_SAMPLING_FREQUENCY_FIELD: 20312.5
        }
    }
)




# ==============================================================================
# Declares the SamplingWindow
# Observation time: 40s
# Sampling time: 20s
# ==============================================================================
sampling_window: SamplingWindow = SamplingWindow(120, 100)

state_queue = multiprocessing.Manager().Queue()

def icomox_test(unique_id: Tuple):
    package_queue_proc_manager = ProcessesManager()
    package_queue_proc_manager.start()
    gateway_packages_queue: PackageQueueProxy = package_queue_proc_manager.PackageQueue()
    icomox_host = IcomoxDriversHost(gateway_packages_queue,state_queue)
    icomox_host.connect()
    icomox_host.connect_drivers()

    icomox_host.add_drivers([(icomox_2_descriptor)])
    task2 = SamplingTask(2,sampling_window, 'client')


    while True:
        try:
            state = state_queue.get(False,10).driver_state
            if state == DriverState.CONNECTED:
                icomox_host.receive_sampling_tasks(
                    {
                        icomox_2_descriptor.logical_id: task2
                    }
                )
                icomox_host.start_sampling_windows()
            else:
                print(state)
        except queue.Empty as ex:
            pass




if __name__ == "__main__":
    """Main function to debug if the module is executed."""
    args = sys.argv[1:]

    unique_id = tuple(args[0].split("-"))
    icomox_test(unique_id)



"""
Exposes a method to retrieve the machines local IP programmatically.
"""
import socket


def get_local_ip() -> str:
    st = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        st.connect(('10.255.255.255', 1))
        IP = st.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        st.close()
    return IP


if __name__ == '__main__':
    add = get_local_ip()
    print(f'Your local IP: {add}')



# class ClassPingPong
# Purpose: Implementing mechanism for a single ping pong buffer
class ClassPingPong:
    def __init__(self):
        self.clear()

    def clear(self):
        self.Buffers = [bytearray(), bytearray()]   # messages that arrive/processed
        self.BufferPopulated = [False, False]       # False = Buffers[] is empty or already have been processed
                                                    # True = Buffers[] is populated and not have been processed yet
        self.SenderBufIndex = 0  #
        self.ReceiverBufIndex = 0  #

    def try_push_msg(self, msg):
        # Here we push the incoming messages from the class_iCOMOX_Communication thread to a ping pong buffer, if possible
        if not self.BufferPopulated[self.SenderBufIndex]:
            self.Buffers[self.SenderBufIndex] = msg
            self.BufferPopulated[self.SenderBufIndex] = True
            self.SenderBufIndex = 1 - self.SenderBufIndex
            return True  # Succeeded to push a message
        else:
            return False  # No message can be pushed

    def try_pop_msg(self):
        if self.BufferPopulated[self.ReceiverBufIndex]:
            msg = self.Buffers[self.ReceiverBufIndex]
            # self.process_incoming_msg(msg)
            self.BufferPopulated[self.ReceiverBufIndex] = False
            self.ReceiverBufIndex = 1 - self.ReceiverBufIndex
            return msg  # Succeeded to pop a message
        else:
            return None  # No message can be pop


class ClassPingPongArr:
    def __init__(self, count):
        self.__PingPongArr__ = []
        for i in range(0, count):
            self.__PingPongArr__.append(ClassPingPong())

    def assert_index(self, index):
        assert ((index >= 0) and (index < len(self.__PingPongArr__)))

    def clear(self, index):
        self.assert_index(index=index)
        self.__PingPongArr__[index].clear()

    def clear_all(self):
        for PingPong in self.__PingPongArr__:
            PingPong.clear()

    def try_push_msg(self, index, msg):
        self.assert_index(index=index)
        return self.__PingPongArr__[index].try_push_msg(msg=msg)

    def try_pop_msg(self, index):
        self.assert_index(index=index)
        return self.__PingPongArr__[index].try_pop_msg()

    def count(self):
        return len(self.__PingPongArr__)



"""
This module currently defines the mocked streamlit application
for a Gateway.

The application view is organized in "Panels". The user can change
between Panels by using the Sidebar and the radio button for each
Panel.
When a Panel is selected, then the main view of the webpage will
be composed of the elements defined to be render in that Panel and
in the current UxState.

The current UxState is always available in the streamlit session state.

Each Panel is rendered by a helper function defined in this script, and inside
that function, the appropriate view is rendered according to the UxState.
"""
# =====================import==================================
import streamlit as st
from typing import List

import txp.devices.ux.remote_gateway_client
from txp.devices.gateway_states_enums import *
from txp.devices.ux.configuration_panel import render_configuration_panel
from txp.devices.ux.devices_paring_panel import render_devices_pairing
from txp.devices.ux.package_collection_panel import render_package_collection
from txp.devices.ux.render_fatal_error import render_fatal_error
import txp.devices.ux.ux_states_protocol as ux_states
from txp.devices.ux.remote_gateway_client import remote_gateway_client
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# =====================Data==================================

# Buttons labels for the Sidebar views
CONFIGURATION_SIDEBAR_BUTTON_LABEL = "Configuración"
DEVICE_PAIRING_SIDEBAR_BUTTON_LABEL = "Emparejar Dispositivos"
PACKAGE_COLLECTION_SIDEBAR_BUTTON_LABEL = "Recolección de Paquetes"

BUTTON_EMOJI_DICT = {
    CONFIGURATION_SIDEBAR_BUTTON_LABEL: "",
    DEVICE_PAIRING_SIDEBAR_BUTTON_LABEL: "",
    PACKAGE_COLLECTION_SIDEBAR_BUTTON_LABEL: ""
}

# =====================Initialize app==================================
def initialize_app():
    """Initializes the app in the first run.

    TODO. When integration with the Gateway, here the initial state/ recovered state
        should be assigned.
    """

    if "initialized" not in st.session_state:
        st.session_state.gateway_state = (
            remote_gateway_client.last_known_state
        )
        remote_gateway_client.start_refresh_loop()

        # Flag to track the sidebar index for the default radio button
        st.session_state.sidebar_radio_index = 0
        st.session_state.configuration_view_first_time = True
        st.session_state.devices_pairing_view_first_time = True
        st.session_state.package_collection_first_time = True

        st.session_state.initialized = True


# ==================== Sidebar ==========================================
def render_sidebar():
    """Render the sidebar of the application. The sidebar allows to change
    between Panels."""
    gateway_state = st.session_state.gateway_state

    # UX Valid views based on the current state.
    # The list of Ux Valid Views is the list of selectable views to show in the sidebar.
    # By design, if a Gateway production process has an error, it'll be replaced.
    # Then, it's safe to add valid views based on the CURRENT state.
    ux_valid_views: List[str] = []

    if isinstance(
        gateway_state,
        (
            ux_states.UxStartupState,
            ux_states.UxConfigurationReceptionState,
            ux_states.UxConfigurationValidationState,
        ),
    ):
        ux_valid_views.append(CONFIGURATION_SIDEBAR_BUTTON_LABEL)
        BUTTON_EMOJI_DICT[CONFIGURATION_SIDEBAR_BUTTON_LABEL] = '⬅️'
        if st.session_state.configuration_view_first_time:
            st.session_state.sidebar_radio_index = 0
            st.session_state.configuration_view_first_time = False

    elif isinstance(gateway_state, ux_states.UxDevicePairingState):
        ux_valid_views.extend(
            [CONFIGURATION_SIDEBAR_BUTTON_LABEL, DEVICE_PAIRING_SIDEBAR_BUTTON_LABEL]
        )
        BUTTON_EMOJI_DICT[CONFIGURATION_SIDEBAR_BUTTON_LABEL] = '✅'
        BUTTON_EMOJI_DICT[DEVICE_PAIRING_SIDEBAR_BUTTON_LABEL] = '⬅️'
        if st.session_state.devices_pairing_view_first_time:
            st.session_state.sidebar_radio_index = 1
            st.session_state.devices_pairing_view_first_time = False

    elif isinstance(gateway_state,
                    (ux_states.UxReadyState, ux_states.UxSamplingState, ux_states.UxPreparingDevicesState)):
        ux_valid_views.extend(
            [
                CONFIGURATION_SIDEBAR_BUTTON_LABEL,
                DEVICE_PAIRING_SIDEBAR_BUTTON_LABEL,
                PACKAGE_COLLECTION_SIDEBAR_BUTTON_LABEL,
            ]
        )
        BUTTON_EMOJI_DICT[CONFIGURATION_SIDEBAR_BUTTON_LABEL] = '✅'
        BUTTON_EMOJI_DICT[DEVICE_PAIRING_SIDEBAR_BUTTON_LABEL] = '✅'
        BUTTON_EMOJI_DICT[PACKAGE_COLLECTION_SIDEBAR_BUTTON_LABEL] = '⬅️'
        if st.session_state.package_collection_first_time:
            st.session_state.sidebar_radio_index = 2
            st.session_state.devices_pairing_view_first_time = False

    st.sidebar.subheader("Estado del Gateway")
    if gateway_state.state == GatewayStates.Ready:
        st.sidebar.markdown(f"{gateway_state.state}-{str(gateway_state.status)}")
    else:
        st.sidebar.markdown(f"{gateway_state.state}")
    st.sidebar.markdown("---------------")

    st.sidebar.markdown("**Pasos**")

    st.sidebar.radio(
        label="Seleccionar Panel", options=ux_valid_views, key="sidebar_panel_radio",
        format_func=lambda label: label + ' ' + BUTTON_EMOJI_DICT[label],
        index=st.session_state.sidebar_radio_index,
    )

# =====================App declarative flow==================================

# This snippet of html will hide the streamlit burguer button
hide_menu_style = """
        <style>
        #MainMenu {visibility: hidden;}
        </style>
        """
st.markdown(hide_menu_style, unsafe_allow_html=True)

initialize_app()
render_sidebar()

if st.session_state.gateway_state.state == GatewayStates.Error:
    if st.session_state.gateway_state.status == GatewayErrorStatus.ProductionNotCompleted:
        render_fatal_error()

elif st.session_state.sidebar_panel_radio == "Configuración":
    render_configuration_panel()

elif st.session_state.sidebar_panel_radio == "Emparejar Dispositivos":
    render_devices_pairing()

elif st.session_state.sidebar_panel_radio == "Recolección de Paquetes":
    render_package_collection()



import struct
import datetime
import os
import sys
import ipaddress


def u16s_to_str(arr, seperator, prefix=""):
    s = ""
    if len(arr) > 0:
        if 1 == (len(arr) % 2):
            arr += b"\x00"
        u16s = struct.unpack("<{:d}H".format(len(arr) / 2))
        for i in range(0, (len(arr) / 2) - 2):
            s += prefix + "{:04X}".format(u16s[i]) + seperator
        s += prefix + "{:04X}".format(u16s[-1])  # add the last element of u16s
    return s


def u8s_to_str(arr, separator, prefix=""):
    s = ""
    if len(arr) > 0:
        for i in range(0, len(arr) - 1):
            s += prefix + "{:02X}".format(arr[i]) + separator
        s += prefix + "{:02X}".format(arr[-1])  # add the last element of arr
    return s


def string_without(str, chars_to_remove):
    return str.translate({ord(c): None for c in chars_to_remove})


def bytearrayToString(bytes):
    index = bytes.find(b"\xFF")
    if index >= 0:
        bytes = bytes[:index]

    index = bytes.find(b"\x00")
    if index >= 0:
        bytes = bytes[:index]

    try:
        result = str(bytes, "utf8")
    except:
        result = ""
    finally:
        pass
    return result


def stringToIpAddress(ipaddress_str):
    try:
        ipaddress_obj = ipaddress.ip_address(ipaddress_str)
    except:
        ipaddress_obj = None
    finally:
        pass
    return ipaddress_obj


def OUT(outStr):
    s = str(datetime.datetime.now()) + " " + outStr
    print(s)


def resource_path(relative_path):
    try:
        # base_path = sys._MEIPASS
        base_path = getattr(sys, '_MEIPASS', os.path.dirname(os.path.abspath(__file__)))
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)


"""
This module contains the concrete implementation case of Driver for the
Icomox POE device.
"""

# ============================ imports =========================================
import queue
import threading
from typing import Tuple, List

from txp.devices.drivers.driver import Driver, EdgeParameterNotFoundError, DriverState
from txp.devices.drivers.icomox.icomox_datahandling import getIcomoxMessageType, getSignalSamples, IcomoxMessageType, \
    IcomoxSensorMassageType, sendSetConfigurationCommand
from txp.common.edge import EdgeDescriptor
from txp.devices.package_queue import PackageQueueProxy
from txp.common.edge import VibrationSpeedSignal, VibrationAccelerationSignal, MicrophoneSignal, MagnetometerSignal, TemperatureSignal, \
    Signal

from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# Icomox driver implementation
# ==============================================================================
class IcomoxDriver(Driver):
    """Implementation of the Icomox POE driver.

    The Icomox Driver is a logical unit configured in the IoT management.
    Operational physical units are mapped to the logical configuration by using
    a Logical ID -> Mac Address mapping. This allows the system to rotate and re assign
    physical sensors without problems.
    """

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            gateway_packages_queue: PackageQueueProxy,
            packages_from_host_queue: queue.PriorityQueue,
            on_driver_state_changed: callable,
            host_clock_notification_queue: queue.Queue
    ):
        """
        Note: This Driver receives a MAC address, which is added and passed down by the Gateway operative
        configuration. The MAC address is not known by the Cloud configuration.

        Args
            edge_descriptor: The Edge Descriptor for this Driver.
            gateway_packages_queue: The Gateway Package Queue to put produced packages.
            smartmesh_manager: The SmartMesh manager that interact with the physical device.
        """
        Driver.__init__(self, edge_descriptor, gateway_packages_queue,
                        on_driver_state_changed, host_clock_notification_queue)
        self._parameters_validation(edge_descriptor)
        self._setup_signals()
        """
        _signal_shape: Internal attribute to control the VibrationSignal dimensions.
        """
        self._edge_packets_queue: queue.PriorityQueue = packages_from_host_queue
        """
        _samples_queue: A queue provided to the listener to handle packages pending to be processed
        with the timestamp as priority.
        """

        self.setup_edge()

        self._lock = threading.Lock()

    def _parameters_validation(self, edge_descriptor):
        self._vibration_adxl356_sensor_parameters_validation(edge_descriptor)
        self._vibration_adxl362_sensor_parameters_validation(edge_descriptor)
        self._magnetometer_bmm150_sensor_parameters_validation(edge_descriptor)
        self._microphone_IM69D130_sensor_parameters_validation(edge_descriptor)

    def _vibration_adxl362_sensor_parameters_validation(self, edge_descriptor):
        if (self.PERCEPTION_DIMENSIONS_FIELD not
                in edge_descriptor.perceptions[VibrationSpeedSignal.perception_name()]):
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received"
                f"in {VibrationAccelerationSignal.perception_name()} perception configuration."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received"
                f"in {VibrationAccelerationSignal.perception_name()} perception configuration."
            )
        if len(self.perceptions[VibrationSpeedSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD]) != 2:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} for {VibrationSpeedSignal.perception_name()}"
                f"should be of length 2"
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} for {VibrationSpeedSignal.perception_name()}"
                f"should be of length 2"
            )
        if self.PERCEPTION_SAMPLING_FREQUENCY_FIELD not in self.perceptions[VibrationSpeedSignal.perception_name()]:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_SAMPLING_FREQUENCY_FIELD} not received "
                f"in {VibrationSpeedSignal.perception_name()} perception configurtion."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_SAMPLING_FREQUENCY_FIELD} not received "
                f"in {VibrationSpeedSignal.perception_name()} perception configurtion."
            )

    def _build_vibration_signal_adlx356_sensor_from_samples(self,samples, sampling_frequency=3611.11, sampling_resolution=16):
        signal_shape: Tuple[int, int] = (
            int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1])
        )
        return VibrationAccelerationSignal.build_signal_from_lists(
            samples,
            signal_shape,
            sampling_frequency,
            sampling_resolution,
        )

    def _build_vibration_signal_adlx362_sensor_from_samples(self,samples, sampling_frequency=399.60, sampling_resolution=16):
        signal_shape: Tuple[int, int] = (
            int(self.perceptions[VibrationSpeedSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[VibrationSpeedSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1])
        )
        return VibrationSpeedSignal.build_signal_from_lists(
            samples,
            signal_shape,
            sampling_frequency,
            sampling_resolution,
        )

    def _build_magnetometer_signal_bmm150_sensor_from_samples(self,samples, sampling_frequency=334.36, sampling_resolution=16):
        signal_shape: Tuple[int, int] = (
            int(self.perceptions[MagnetometerSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[MagnetometerSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1]),
        )
        return MagnetometerSignal.build_signal_from_lists(
            samples,
            signal_shape,
            sampling_frequency,
            sampling_resolution,
        )

    def _build_microphone_signal_im69d130_sensor_from_samples(self,samples, sampling_frequency=20312, sampling_resolution=16):
        signal_shape: Tuple[int, int] = (
            int(self.perceptions[MicrophoneSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[MicrophoneSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1]),
        )
        return MicrophoneSignal.build_signal_from_lists(
            samples,
            signal_shape,
            sampling_frequency,
            sampling_resolution,
        )

    def _build_temperature_signal_adt7410_sensor_from_samples(self,temperature):
        return TemperatureSignal.build_signal_from_value(temperature, signal_frequency=None)

    @classmethod
    def device_name(cls) -> str:
        return "Icomox"

    @classmethod
    def is_virtual(cls) -> bool:
        return False

    def observation_pulse(self) -> int:
        """The Icomox device can sample from each one of its perspectives each second.

        As a consequence, 12 seconds is enough time to collect samples from all the perspectives.
        """
        return 10

    def setup_edge(self):
        pass

    def connect_device(self):
        """
        All connection states is handled by the icomox_driver_host module for the nature of the communication
        it is always initiated by the icomox client via tcp ip. driver never initiates the communication.
        """
        pass

    def disconnect_device(self):
        """
        All connection states is handled by the icomox_driver_host module for the nature of the communication
        it is always initiated by the icomox client via tcp ip. driver never initiates the communication.
        """
        pass

    def _flush_packets_queue(self):
        while not self._edge_packets_queue.empty():
            self._edge_packets_queue.get()

    def _start_sampling_collect(
            self
    ) -> List["Signal"]:
        """This method will collect the samples obtained from the edge device and will
        produce the system Signals to send the package to the gateway.

        This method should collect samples that belong to the defined Sampling Time.

        Args:
            sampling_time_finished: event to check when the sampling time has finished.
        Returns:
            A list of Signals collected by the Icomox.
        """

        signals = []

        transmiting = False
        received_sensors = set()
        self._flush_packets_queue()
        while True:
            try:
                if not transmiting:
                    sendSetConfigurationCommand(self.get_complete_device_id().physical_id,True)
                    transmiting = True
                timeout_in_sec = 0.5
                msg = self._edge_packets_queue.get(True, timeout_in_sec)[1]
                log.debug("Message for Icomox POE received. UniqueId: {}".format(self.get_complete_device_id().physical_id))
                message_type = getIcomoxMessageType(msg)
                if message_type == IcomoxMessageType.REPORT_MESSAGE:
                    sensor_type = self._process_report_message(msg, signals)
                    received_sensors.add(sensor_type)
                elif message_type == IcomoxMessageType.API_MESSAGE:
                    self._process_api_message(msg)

            except queue.Empty:
                pass
            except Exception as ex:
                pass

            if len(received_sensors.intersection({
                IcomoxSensorMassageType.VIBRATION_SENSOR_ADXL356,
                IcomoxSensorMassageType.VIBRATION_SENSOR_ADXL362,
                IcomoxSensorMassageType.MAGNETOMETER_SENSOR_BMM150,
                IcomoxSensorMassageType.MICROPHONE_SENSOR_IM69D130,
                IcomoxSensorMassageType.TEMPERATURE_SENSOR_ADT7410
            })) == 5:
                return signals

            if self._stop_current_pulse_signal.is_set():
                return []

    def _stop_sampling_collect(self):
        """This method will stop the sampling collection in the driver implementation.
        """
        if self.driver_state == DriverState.CONNECTED or self.driver_state == DriverState.SAMPLING:
            sendSetConfigurationCommand(self.get_complete_device_id().physical_id,False)
        else:
            log.warning("{} state is {}. couldn't turn of data sending".format(self.logical_id, self.driver_state))


    def _process_report_message(self, msg, signals) -> IcomoxSensorMassageType:
        signalType, samples = getSignalSamples(msg)
        log.debug("driver {} signal received {} and message lenght is {} samples lenth is {}".format(self.device_name(),signalType,len(msg),len(samples)))
        if signalType == IcomoxSensorMassageType.VIBRATION_SENSOR_ADXL362:
            self._process_vibration_signal_adlx362(samples, signals)
        elif signalType == IcomoxSensorMassageType.VIBRATION_SENSOR_ADXL356:
            self._process_vibration_signal_adlx356(samples, signals)
        elif signalType == IcomoxSensorMassageType.MAGNETOMETER_SENSOR_BMM150:
            self._process_magnetometer_signal_bmm150(samples, signals)
        elif signalType == IcomoxSensorMassageType.MICROPHONE_SENSOR_IM69D130:
            self._process_microphone_signal_im69d130(samples, signals)
        elif signalType == IcomoxSensorMassageType.TEMPERATURE_SENSOR_ADT7410:
            self._process_temperature_signal_adt7410(samples, signals)

        log.debug(f"Icomox {self.logical_id} received a signal {signalType.name}")

        return signalType

    def _process_temperature_signal_adt7410(self, samples, signals):
        if samples is not None:
            signal = TemperatureSignal.build_signal_from_value(
                samples[0],
                0
            )
            signals.append(signal)
        else:
            log.error('Icomox RAW data is corrupted')

    def _process_microphone_signal_im69d130(self, samples, signals):
        if samples is not None:
            signal = MicrophoneSignal.build_signal_from_lists(
                samples,
                self._signal_shape_im69d130,
                self._sampling_frequency_im69d130,
                16,
            )
            signals.append(signal)
        else:
            log.error('Icomox RAW data is corrupted')

    def _process_magnetometer_signal_bmm150(self, samples, signals):
        if samples is not None:
            signal = MagnetometerSignal.build_signal_from_lists(
                samples,
                self._signal_shape_bmm150,
                self._sampling_frequency_bmm150,
                16,
            )
            signals.append(signal)
        else:
            log.error('Icomox RAW data is corrupted')

    def _process_vibration_signal_adlx356(self, samples, signals):
        if samples is not None:
            signal = VibrationAccelerationSignal.build_signal_from_lists(
                samples,
                self._signal_shape_adlx356,
                self._sampling_frequency_adlx356,
                16,
            )
            signals.append(signal)
        else:
            log.error('Icomox RAW data is corrupted')

    def _process_vibration_signal_adlx362(self, samples, signals):
        if samples is not None:
            signal = VibrationSpeedSignal.build_signal_from_lists(
                samples,
                self._signal_shape_adlx362,
                self._sampling_frequency_adlx362,
                16,
            )
            signals.append(signal)
        else:
            log.error('Icomox RAW data is corrupted')

    def _magnetometer_bmm150_sensor_parameters_validation(self, edge_descriptor):
        if self.PERCEPTION_DIMENSIONS_FIELD not in self.perceptions[MagnetometerSignal.perception_name()]:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"{self.perceptions[MagnetometerSignal.perception_name()]} perception configuration"
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"{self.perceptions[MagnetometerSignal.perception_name()]} perception configuration"
            )
        if len(self.perceptions[MagnetometerSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD]) != 2:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} for {MagnetometerSignal.perception_name()} should"
                f"have length of 2"
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} for {MagnetometerSignal.perception_name()} should"
                f"have length of 2"
            )
        if self.PERCEPTION_SAMPLING_FREQUENCY_FIELD not in self.perceptions[MagnetometerSignal.perception_name()]:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"in {self.perceptions[MagnetometerSignal.perception_name()]} perception configuration."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"in {self.perceptions[MagnetometerSignal.perception_name()]} perception configuration."
            )

    def _microphone_IM69D130_sensor_parameters_validation(self, edge_descriptor):
        if self.PERCEPTION_DIMENSIONS_FIELD not in self.perceptions[MicrophoneSignal.perception_name()]:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"in {MicrophoneSignal.perception_name()} perception configuration."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"in {MicrophoneSignal.perception_name()} perception configuration."
            )
        if len(self.perceptions[MicrophoneSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD]) != 2:
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} for {MicrophoneSignal.perception_name()} "
                f"should have a length of 2."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} for {MicrophoneSignal.perception_name()} "
                f"should have a length of 2."
            )

    def _vibration_adxl356_sensor_parameters_validation(self, edge_descriptor):
        if (self.PERCEPTION_DIMENSIONS_FIELD not
                in edge_descriptor.perceptions[VibrationAccelerationSignal.perception_name()]):
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"in {VibrationAccelerationSignal.perception_name()} perception."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received "
                f"in {VibrationAccelerationSignal.perception_name()} parameters."
            )
        if (len(edge_descriptor.perceptions[VibrationAccelerationSignal.perception_name()]
               [self.PERCEPTION_DIMENSIONS_FIELD]) != 2):
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} in {VibrationAccelerationSignal.perception_name()} "
                f"perception should be have length 2."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_DIMENSIONS_FIELD} in {VibrationAccelerationSignal.perception_name()} "
                f"perception should be have length 2."
            )
        if (self.PERCEPTION_SAMPLING_FREQUENCY_FIELD not
                in edge_descriptor.perceptions[VibrationAccelerationSignal.perception_name()]):
            log.error(
                f"IcomoxDriver: {self.PERCEPTION_SAMPLING_FREQUENCY_FIELD} not received"
                f"in {VibrationAccelerationSignal.perception_name()} parameters."
            )
            raise EdgeParameterNotFoundError(
                f"IcomoxDriver: {self.PERCEPTION_SAMPLING_FREQUENCY_FIELD} not received"
                f"in {VibrationAccelerationSignal.perception_name()} parameters."
            )

    def _process_api_message(self, msg):
        #TODO: check here which one is and flag the response to stop timer when a command is sent and is waiting for a response
        pass

    def _setup_signals(self):
        self._setup_signal_adlx356()
        self._setup_signal_adlx362()
        self._setup_signal_im69d130()
        self._setup_signal_bmm150()

    def _setup_signal_bmm150(self):
        self._signal_shape_bmm150: Tuple[int, int] = (
            int(self.perceptions[MagnetometerSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[MagnetometerSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1]),
        )
        self._sampling_frequency_bmm150 = (self.perceptions[MagnetometerSignal.perception_name()]
            [self.PERCEPTION_SAMPLING_FREQUENCY_FIELD])

    def _setup_signal_im69d130(self):
        self._signal_shape_im69d130: Tuple[int, int] = (
            int(self.perceptions[MicrophoneSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[MicrophoneSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1])
        )
        self._sampling_frequency_im69d130 = self.perceptions[MicrophoneSignal.perception_name()][self.PERCEPTION_SAMPLING_FREQUENCY_FIELD]

    def _setup_signal_adlx362(self):
        self._signal_shape_adlx362: Tuple[int, int] = (
            int(self.perceptions[VibrationSpeedSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[VibrationSpeedSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1]),
        )
        self._sampling_frequency_adlx362 = (self.perceptions[VibrationSpeedSignal.perception_name()]
        [self.PERCEPTION_SAMPLING_FREQUENCY_FIELD])

    def _setup_signal_adlx356(self):
        self._signal_shape_adlx356: Tuple[int, int] = (
            int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1]),
        )
        self._sampling_frequency_adlx356 = (self.perceptions[VibrationAccelerationSignal.perception_name()]
        [self.PERCEPTION_SAMPLING_FREQUENCY_FIELD])




"""
This module defines the Gateway States-Status related enumerations.
"""
# ============================ imports =========================================
import enum


# ==============================================================================
# Definition of the custom Gateway States enumeration
# ==============================================================================
class GatewayStates(enum.Enum):
    def __str__(self):
        return self.name

    Uninitialized = 0
    """Uninitialized: Internal state to set and keep consistency while the states machine is being 
    programatically initialized"""

    Startup = 1
    """Startup: Receiving initial files and values from the user interaction with the Gateway Local 
    web page."""

    # TODO: Delete this.
    StartupError = 2
    """StartupError: Error from Startup state"""

    ConfigurationReception = 3
    """ConfigurationReception: State in which the cloud logical configuration is received"""

    # TODO: Delete this.
    ConfigurationReceptionError = 4
    """ConfigurationReceptionError: Error from ConfigurationReception state"""

    ConfigurationValidation = 5
    """ConfigurationValidation: State to process the received configuration and choose a path of execution"""

    PreparingDevices = 7
    """PreparingDevices: State to transition when the gateway needs to create the DriversHost and Drivers
    This state can be reached after production phase, or when a produced gateway is turned on.
    """

    DevicesPairing = 8
    """DevicesPairing: State to step into when an edge devices pairing is required"""

    DevicePairingError = 9
    """DevicePairingError: Error from DevicesPairing state"""

    Ready = 10
    """Ready: The Gateway instance is ready to operate"""

    Sampling = 11
    """Sampling: The gateway is currently sampling raw data from the edge devices"""

    Error = 12
    """Error: A General Error state to handle error cases."""


class StartupStatus(enum.Enum):
    RequestRequiredValues = 1
    """Request the required values from the user input."""
    RequiredValuesReceived = 2
    """The user has entered the required values. The Gateway will proceed to test those."""


class ConfigurationReceptionStatus(enum.Enum):
    EstablishCloudConnection = 1
    Idle = 2
    ListenGatewayConfiguration = 3
    ListenMachinesConfiguration = 4
    ListenEdgesConfiguration = 5


class ConfigurationValidationStatus(enum.Enum):
    CheckConfigurations = 1
    UserConfirmedSettings = 2


class PreparingDevicesStatus(enum.Enum):
    CreatingDriversHosts = 1
    CreatingDrivers = 2
    WaitingForDevicesToBeConnected = 3


class ConfigurationUpdateStatus(enum.Enum):
    UpdatePreviousConfiguration = 1


class DevicePairingStatus(enum.Enum):
    SelectDeviceToPair = 1
    PairingSelectedDevice = 2
    RequestConfirm = 3
    PairingConfirmed = 4
    PairingCancelled = 5
    PairingError = 6

    # Virtual Pairing Status
    RequestVirtualCameraConnection = 7
    VirtualCameraConnectionConfirmed = 8
    VirtualCameraProducedSuccessfully = 9
    VirtualCameraConnectionError = 10

    RequestVirtualThermocameraConnection = 11
    VirtualThermocameraConnectionConfirmed = 12
    VirtualThermocameraProducedSuccessfully = 13
    VirtualThermocameraConnectionError = 14

    VirtualRequestNextStepFromUser = 15  # Cameras already paired, the user selects what to do
    VirtualEnableTorque = 16
    VirtualDisableTorque = 17
    VirtualCapturePosition = 18
    VirtualRejectCapturedPosition = 19
    VirtualConfirmCapturedPosition = 20
    VirtualConfirmAllPositions = 21


class DevicePairingErrorStatus(enum.Enum):
    ErrorPairingDevice = 1


class ReadyStatus(enum.Enum):
    def __str__(self):
        return self.name
    Idle = 1
    SchedulingSampling = 2


class GatewaySamplingStatus(enum.Enum):
    def __str__(self):
        return self.name

    Sampling = 1
    CollectPackages = 2
    SendPackages = 3


# =========================== ERROR known error status ==============================
class GatewayErrorStatus(enum.Enum):
    ProductionNotCompleted = 1



"""
Defines a common encoder designed to work with the dataclasses that requires
to be handled in JSON-string format in IoT Core:
    - GatewayConfig
    - EdgeDescriptor
    - SamplingWindow
    - MachineMetadata
"""
import json
import abc
from typing import Dict


class ComplexEncoder(json.JSONEncoder):
    """Our common complex encoder to use when serializing a class instance to JSON.

    Example:
        edge_descriptor = EdgeDescriptor(...)
        js = json.dumps(edge_descriptor_1, cls=ComplexEncoder)

    For nested cases like the GatewayConfig, this encoder will also work recursively
    as long as the nested classes defines the reprJSON() method.

    Example:
        gateway_config = GatewayConfig(
            "gateway-test-proto",
            "registry-test-proto",
            "project-test-proto",
            "region-test-proto",
            [sampling_window],
            [machine_metadata.machine_id]
        )
        json_str = json.dumps(
            gateway_config, cls=ComplexEncoder, indent=4
        )
    """
    def default(self, obj):
        if hasattr(obj, 'reprJSON'):
            return obj.reprJSON()
        else:
            return json.JSONEncoder.default(self, obj)


class ComplexEncoderSerializable(abc.ABC):

    @abc.abstractmethod
    def reprJSON(self) -> Dict:
        """Every child class that wants to use the ComplexEncoder,
        should implement this method."""
        pass



"""
This module implements the Telemetry Health view for the application.

NOTE: This is just a dummy implementation of a MainView in order to shown the
mechanism provided by the `core_components` package.
"""
from txp.web.core_components.app_profiles import AppProfile
from txp.web.core_components.main_view import MainView
import txp.cloud.controllers.firestore as firestore_controller
import logging
import requests
from fastapi import status
from typing import Dict
import streamlit as st
import datetime
import pytz

# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class CommitView(MainView):
    """The Telemetry Health View"""

    def __init__(self):
        super(CommitView, self).__init__(component_key="commit_view")
        self.data = None
        self.machines = None
        self.range_time = {}

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.ANALYTICS

    @classmethod
    def get_display_name(cls) -> str:
        return "Publicar dataset"

    def get_data(self):
        if self.data is None:
            db = firestore_controller.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )

            self.data = firestore_controller.pull_recent_project_data(db, self._tenant_id)
            self.machines = {d["machine_id"]: d for d in self.data["machines"]}
            self.tasks = self.get_tasks_per_machines()
            self.current_machine = list(self.machines.keys())[0]
            self.task = self.tasks[self.current_machine][0]
            self.url = st.secrets["service_url"]
            db.close()

    def get_tasks_per_machines(self):
        tasks_per_machine = {}
        for machine, values in self.machines.items():
            tasks = []
            if type(values["tasks"]) == dict:
                for task in values["tasks"].values():
                    tasks.append(task["task_id"])
            else:
                for task in values["tasks"]:
                    tasks.append(task["task_id"])
            tasks_per_machine.update({machine: tasks})
        return tasks_per_machine

    def _build_submenu_state(self) -> Dict:
        pass

    def _build_content_state(self) -> Dict:
        pass

    def _render_submenu(self):
        pass

    def parsed_versions(self, versions):
        versions_list = versions.replace(" ", "").split(",")
        return versions_list

    def _format_range_date(self):
        tz = pytz.timezone("America/Mexico_City")
        full_dates = []

        full_date_start = datetime.datetime.combine(
            date=self.range_time['start_date'],
            time=self.range_time['start_time']
        )
        full_date_end = datetime.datetime.combine(
            date=self.range_time['end_date'],
            time=self.range_time['end_time']
        )

        for full_date in [full_date_start, full_date_end]:
            localized_full_date: datetime.datetime = tz.localize(full_date)
            utc_full_date: datetime.datetime = localized_full_date.astimezone(pytz.utc)
            full_dates.append(utc_full_date.strftime("%Y-%m-%d %H:%M:%S.%f%z"))

        return full_dates[0], full_dates[1]

    def  _commit(self):
        versions = self.parsed_versions(
            st.session_state[f"{self.__class__.__name__}_dataset_version"]
        )

        self.range_time = {
            'start_date': st.session_state[f"{self.__class__.__name__}_date_range"][0],
            'end_date': st.session_state[f"{self.__class__.__name__}_date_range"][1],
            'start_time': st.session_state[f"{self.__class__.__name__}_start"],
            'end_time': st.session_state[f"{self.__class__.__name__}_end"]
        }

        full_start, full_end = self._format_range_date()

        commit_json = {
            "asset_id": st.session_state[f"{self.__class__.__name__}_machine"],
            "task_id": st.session_state[f"{self.__class__.__name__}_task"],
            "dataset_name": st.session_state[f"{self.__class__.__name__}_dataset_name"],
            "tenant_id": self._tenant_id,
            "dataset_versions": versions,
            "start": full_start,
            "end": full_end,
            "bigquery_dataset": self._bigquery_dataset,
        }
        try:
            headers = {"Authorization": f"Bearer {st.secrets['service_auth_token']}"}
            response = requests.post(f"{self.url}/annotation/commit", json=commit_json, headers=headers)
        except (ConnectionError, ConnectionRefusedError) as ce:
            st.error("No se pudo establecer con el servidor de anotación. "
                     "Contacte a soporte.")
        except Exception as e:
            st.error("No se pudo establecer con el servidor de anotación. "
                     "Contacte a soporte.")
        else:
            json_response = response.json()
            if response.status_code == status.HTTP_201_CREATED:
                st.success(
                    f'El dataset {st.session_state[f"{self.__class__.__name__}_dataset_name"]} versión {str(versions)} inició '
                    f"su publicación satisfactoriamente."
                )
                print(json_response)
            elif response.status_code == status.HTTP_412_PRECONDITION_FAILED:
                st.warning(
                    f'El dataset no se puede generar en este momento, hay anotaciones procesandose. Intente más tarde.'
                )
            else:
                st.error(
                    f'No se pudo publicar el dataset {st.session_state[f"{self.__class__.__name__}_dataset_name"]} versión {str(versions)}.'
                )
                print(f"Unexpected error during commit request {response}, {json_response}")
                return None

    def _render_content(self):
        # helper functions
        self.get_data()
        timezone = pytz.timezone("America/Mexico_City")
        st.header("Publicar dataset de entrenamiento")

        m = st.markdown(
            """
                               <style>
                               div.stButton > button:first-child {
                               color: #0000CD;
                               box-sizing: 5%;
                               height: 2em;
                               width: 100%;
                               font-size:20px;
                               border: 2px solid;
                               border-color: #D3D3D3;
                               border-radius: 10px;
                               box-shadow: 1px 1px 1px 1px 
                               rgba(0, 0, 0, 0.1);
                               padding-top: 0px;
                               padding-bottom: 0px;
                               }
                               .center {
                               margin-left: auto;
                               margin-right: auto;
                               }
                               </style>""",
            unsafe_allow_html=True,
        )

        st.selectbox(
            "Indique el activo:",
            options=list(self.machines.keys()),
            key=f"{self.__class__.__name__}_machine",
            index=0,
        )

        tasks = self.tasks[st.session_state[f"{self.__class__.__name__}_machine"]]

        with st.form("commit"):
            st.selectbox(
                "Escoja un task:",
                options=tasks,
                key=f"{self.__class__.__name__}_task",
                index=0,
            )

            st.date_input(
                "Inserte un periodo:",
                key=f"{self.__class__.__name__}_date_range",
                value=(
                    datetime.datetime.today()
                    if "init_date" not in self.range_time
                    else self.range_time["start_date"],
                    datetime.datetime.today()
                    if "finish_date" not in self.range_time
                    else self.range_time["end_date"]
                ),
            )
            st.time_input(
                "Hora de inicio de la data:",
                key=f"{self.__class__.__name__}_start",
                value=datetime.datetime.now().astimezone(tz=timezone)
                    if "start_time" not in self.range_time
                    else self.range_time["start_time"],
            )
            st.time_input(
                "Hora de culminación de la data:",
                key=f"{self.__class__.__name__}_end",
                value=datetime.datetime.now().astimezone(tz=timezone)
                    if "end_time" not in self.range_time
                    else self.range_time["end_time"]
            )

            st.text_input(
                "Inserte nombre de dataset:",
                key=f"{self.__class__.__name__}_dataset_name",
            )

            st.text_input(
                "Inserte el nombre de la versión. Si quiere especificar más de una version, ingrese los enunciados separados por coma:",
                key=f"{self.__class__.__name__}_dataset_version",
            )
            st.form_submit_button("Publicar dataset", on_click=self._commit)



import numpy as np
from common import iCOMOX_messages
from scipy.signal import welch
from scipy.signal.windows import hamming
import struct

class class_BMM150:
    def __init__(self, quantile_for_noise_floor_estimator, minimum_SNR_for_speed_detection_dB):
        TicksInSamplingPeriod = int(0.5 + (0.003 / (1/32768)))  # number of ticks in 3msec
        self.Fs_Mag = 32768 / TicksInSamplingPeriod # 1/0.003
        self.NFFT_Mag = 512
        self.epsilon = 1E-100
        self.quantile_for_noise_floor_estimation = quantile_for_noise_floor_estimator
        self.minimum_SNR_for_speed_detection = pow(10, minimum_SNR_for_speed_detection_dB/10)      # default threshold of minimum of 20dB above the estimated noise floor density

    def to_PSD(self, mag_x_uT, mag_y_uT, mag_z_uT):
        (freq_mag, Pxxf_Mag_X) = welch(mag_x_uT, self.Fs_Mag, nfft=self.NFFT_Mag, return_onesided=True, detrend=False)
        (freq_mag, Pxxf_Mag_Y) = welch(mag_y_uT, self.Fs_Mag, nfft=self.NFFT_Mag, return_onesided=True, detrend=False)
        (freq_mag, Pxxf_Mag_Z) = welch(mag_z_uT, self.Fs_Mag, nfft=self.NFFT_Mag, return_onesided=True, detrend=False)
        return freq_mag, Pxxf_Mag_X, Pxxf_Mag_Y, Pxxf_Mag_Z

    def readings_to_uTesla(self, raw_samples):
        # Convert the reading to uTesla (by dividing by 16)
        # axis_len = len(raw_samples) // 3
        # mag_x_uTesla = np.array(raw_samples[0:axis_len]) / 16
        # mag_y_uTesla = np.array(raw_samples[axis_len:2*axis_len]) / 16
        # mag_z_uTesla = np.array(raw_samples[2*axis_len:]) / 16
        mag_x_uTesla = np.array(raw_samples[0::3]) / 16
        mag_y_uTesla = np.array(raw_samples[1::3]) / 16
        mag_z_uTesla = np.array(raw_samples[2::3]) / 16
        return mag_x_uTesla, mag_y_uTesla, mag_z_uTesla

    def msg_bytes_to_samples_uTesla(self, payload):
        samples = struct.unpack("<{}h".format(iCOMOX_messages.BMM150_SAMPLES_NUM * 3), payload)
        return self.readings_to_uTesla(raw_samples=samples)

    # Expecting Pxxf values to NOT be in dB. the DC component is not taken into account
    def maximum_of_PSD(self, time_series):
        (freq, Pxxf) = welch(time_series - np.mean(time_series), self.Fs_Mag, nfft=self.NFFT_Mag, return_onesided=True, detrend=False)
        noise_level = np.quantile(Pxxf, self.quantile_for_noise_floor_estimation)
        index_of_max_energy = np.argmax(Pxxf[1:-1]) + 1
        max_energy = Pxxf[index_of_max_energy]
        if max_energy/noise_level >= self.minimum_SNR_for_speed_detection:
            return self.parabola_peak(x1=freq[index_of_max_energy], dx=freq[1], y0=Pxxf[index_of_max_energy-1], y1=Pxxf[index_of_max_energy], y2=Pxxf[index_of_max_energy+1])
            # return freq[index_of_max_energy]
        else:
            return 0.0

    def parabola_peak(self, x1, dx, y0, y1, y2):
        return x1 - (y2 - y0) * dx / (2 * (y2 - 2 * y1 + y0))